#!/usr/bin/env python
#
# glvolume.py - The GLVolume class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`GLVolume` class, which creates and
encapsulates the data and logic required to render 2D slice of an
:class:`.Image` instance.
"""


import logging

import numpy                     as np
import OpenGL.GL                 as gl

import fsl.utils.idle            as idle
import fsl.transform.affine      as affine
import fsleyes.gl                as fslgl
import fsleyes.gl.routines       as glroutines
import fsleyes.gl.shaders.filter as glfilter
from . import                       textures
from . import                       glimageobject
from . import resources          as glresources


log = logging.getLogger(__name__)


class GLVolume(glimageobject.GLImageObject):
    """The ``GLVolume`` class is a :class:`.GLImageObject` which encapsulates
    the data and logic required to render  :class:`.Image` overlays in 2D and
    3D.


    A ``GLVolume`` instance may be used to render an :class:`.Image` instance
    which has a :attr:`.Display.overlayType` equal to ``volume``. It is
    assumed that this ``Image`` instance is associated with a
    :class:`.Display` instance which, in turn, contains a :class:`.VolumeOpts`
    instance, containing display options specific to volume rendering.


    **Version dependent modules**


    The ``GLVolume`` class makes use of the functions defined in the
    :mod:`.gl14.glvolume_funcs` or the :mod:`.gl21.glvolume_funcs` modules,
    which provide OpenGL version specific details for creation/storage of
    vertex data, and for rendering.


    These version dependent modules must provide the following functions:

    ===================================== =====================================
    ``init(GLVolume)``                    Perform any necessary initialisation.

    ``destroy(GLVolume)``                 Perform any necessary clean up.

    ``compileShaders(GLVolume)``          (Re-)Compile the shader programs.

    ``updateShaderState(GLVolume)``       Updates the shader program states
                                          when display parameters are changed.

    ``preDraw(GLVolume)``                 Initialise the GL state, ready for
                                          drawing.

    ``draw2D(GLVolume, zpos, xform)``     Draw a slice of the image at the
                                          given ``zpos``. If ``xform`` is not
                                          ``None``, it must be applied as a
                                          transformation on the vertex
                                          coordinates.

    ``draw3D(GLVolume, xform)``           Draw the image in 3D. If ``xform``
                                          is not ``None``, it must be applied
                                          as a transformation on the vertex
                                          coordinates.

    ``drawAll(Glvolume, zposes, xforms)`` Draws slices at each of the
                                          specified ``zposes``, applying the
                                          corresponding ``xforms`` to each.

    ``postDraw(GLVolume)``                Clear the GL state after drawing.
    ===================================== =====================================


    **2D rendering**


    Images are rendered in essentially the same way, regardless of which
    OpenGL version-specific module is used.  The image data itself is stored
    as an :class:`.ImageTexture`. This ``ImageTexture`` is managed by the
    :mod:`.resources` module, so may be shared by many ``GLVolume`` instances.
    The current colour maps (defined by the :attr:`.VolumeOpts.cmap` and
    :attr:`.VolumeOpts.negativeCmap` properties) are stored as
    :class:`.ColourMapTexture` instances.  A slice through the texture is
    rendered using six vertices, located at the respective corners of the
    image bounds.


    Image voxels may be clipped according to the
    :attr:`.VolumeOpts.clippingRange` property. By default, the voxel value
    is compared against the clipping range, but the
    :attr:`.VolumeOpts.clipImage` property allows the data from a different
    image (of the same dimensions) to be used for clipping. If specified,
    this clipping image is stored as another :class:`.ImageTexture`.


    If the :attr:`.ColourMapOpts.modulateAlpha` setting is active, the opacity
    of rendered voxels is modulated by the voxel intensity.  If
    :attr:`.VolumeOpts.modulateImage` is also set, the opacity is modulated by
    the values in a different image - in this case, this image is stored in
    another :class:`.ImageTexture`.

    An :class:`.AuxImageTextureManager` is used to manage the clip and modulate
    textures.


    **3D rendering**


    In 3D, images are rendered using a ray-casting approach. The image
    bounding box is drawn as a cuboid. Then for each pixel, a ray is cast
    from the 'camera' through the image texture. The resulting colour
    is generated from sampling points along the ray.

    The ``glvolume_funcs`` modules are expected to perform this rendering
    off-screen, using two :class:`.RenderTexture` instances, available as
    attributes ``renderTexture1`` and``renderTexture2``.  After a call to
    ``draw3D``, the final result is assuemd to be contained in
    ``renderTexture1``.


    **Textures**


    The ``GLVolume`` class uses the following textures:

     - An :class:`.ImageTexture`, a 3D texture which contains image data.
       This is bound to texture unit 0.

     - A :class:`.ColourMapTexture`, a 1D texture which contains the
       colour map defined by the :attr:`.VolumeOpts.cmap` property.
       This is bound to texture unit 1.

     - A :class:`.ColourMapTexture`, a 1D texture which contains the
       colour map defined by the :attr:`.VolumeOpts.negativeCmap` property.
       This is bound to texture unit 2.

     - An :class:`.ImageTexture` which contains the clippimg image data.
       This is bound to texture unit 3. If the :attr:`.VolumeOpts.clipImage`
       property is not specified (i.e. it has a value of ``None``), this
       texture will not be bound - in this case, the image texture is used
       for clipping.

     - Two :class:`.RenderTexture` instances which are used for 3D rendering.
       Both of these textures have depth buffers. When one of these textures
       is being drawn it is bound to texture units 4 (for RGBA) and 5 (for
       depth).


    **Attributes**


    The following attributes are available on a ``GLVolume`` instance:

    ==================== ==================================================
    ``imageTexture``     The :class:`.ImageTexture` which stores the image
                         data.
    ``clipTexture``      The :class:`.ImageTexture` which stores the clip
                         image data.
    ``modulateTexture``  The :class:`.ImageTexture` which stores the
                         modulate image data.
    ``colourTexture``    The :class:`.ColourMapTexture` used to store the
                         colour map.
    ``negColourTexture`` The :class:`.ColourMapTexture` used to store the
                         negative colour map.
    ``renderTexture1``   The first :class:`.RenderTexture` used for 3D
                         rendering.
    ``renderTexture2``   The first :class:`.RenderTexture` used for 3D
                         rendering.
    ``texName``          A name used for the ``imageTexture``,
                         ``colourTexture``, and ``negColourTexture`. The
                         name for the latter is suffixed with ``'_neg'``.
    ==================== ==================================================
    """


    def __init__(self, image, overlayList, displayCtx, threedee):
        """Create a ``GLVolume`` object.

        :arg image:       An :class:`.Image` object.

        :arg overlayList: The :class:`.OverlayList`

        :arg displayCtx:  The :class:`.DisplayContext` object managing the
                          scene.

        :arg threedee:    Set up for 2D or 3D rendering.
        """

        glimageobject.GLImageObject.__init__(self,
                                             image,
                                             overlayList,
                                             displayCtx,
                                             threedee)

        # Create an image texture, clip texture, and a colour map texture
        #
        # We use the gl.resources module to manage texture
        # creation, because ImageTexture instances can
        # potentially be shared between GLVolumes. So all
        # GLVolumes use the same name, defined here, to
        # refer to the ImageTexture for a given image.
        self.texName  = '{}_{}'.format(type(self).__name__, id(self.image))

        # Ref to an OpenGL shader program -
        # the glvolume_funcs module will
        # create this for us.
        self.shader = None

        # aux texture manager takes care
        # of clip+modulate textures
        self.auxmgr = textures.AuxImageTextureManager(
            self, clip=None, modulate=None)
        self.auxmgr.registerAuxImage('clip',     self.opts.clipImage)
        self.auxmgr.registerAuxImage('modulate', self.opts.modulateImage)

        # Refs to all of the texture objects.
        self.imageTexture     = None
        self.colourTexture    = textures.ColourMapTexture(self.texName)
        self.negColourTexture = textures.ColourMapTexture(
            '{}_neg'.format(self.texName))

        if self.threedee:

            self.smoothFilter = glfilter.Filter('smooth', texture=0)
            self.smoothFilter.set(kernSize=self.opts.smoothing * 2)

            self.renderTexture1 = textures.RenderTexture(
                self.name, interp=gl.GL_LINEAR, rttype='cd')
            self.renderTexture2 = textures.RenderTexture(
                self.name, interp=gl.GL_LINEAR, rttype='cd')

        # This attribute is used by the
        # updateShaderState method to
        # make sure that the Notifier.notify()
        # method gets called when needed.
        # See that method for details.
        self.__alwaysNotify = False

        self.refreshColourTextures()
        self.refreshImageTexture()

        # Add listeners to this image so the view can be
        # updated when its display properties are changed
        self.addDisplayListeners()

        # Call glvolume_funcs.init when the image
        # and clip textures are ready to be used.
        def init():

            # In some crazy circumstances,
            # a just-created GLVolume can
            # get destroyed immediately
            if not self.destroyed:
                fslgl.glvolume_funcs.init(self)
                self.notify()

        idle.idleWhen(init, self.texturesReady)


    def destroy(self):
        """This must be called when this :class:`GLVolume` object is no
        longer needed. It performs any needed clean up of OpenGL data (e.g.
        deleting texture handles), calls :meth:`removeDisplayListeners`,
        and calls :meth:`.GLImageObject.destroy`.
        """

        self.removeDisplayListeners()

        self.imageTexture.deregister(self.name)
        glresources.delete(self.imageTexture.name)

        self.colourTexture   .destroy()
        self.negColourTexture.destroy()
        self.auxmgr          .destroy()

        self.auxmgr           = None
        self.imageTexture     = None
        self.colourTexture    = None
        self.negColourTexture = None

        if self.threedee:
            self.renderTexture1.destroy()
            self.renderTexture2.destroy()
            self.smoothFilter  .destroy()
            self.renderTexture1 = None
            self.renderTexture2 = None
            self.smoothFilter   = None

        fslgl.glvolume_funcs       .destroy(self)
        glimageobject.GLImageObject.destroy(self)


    def ready(self):
        """Returns ``True`` if this ``GLVolume`` is ready to be drawn,
        ``False`` otherwise.
        """
        return (not self.destroyed      and
                self.shader is not None and
                self.texturesReady())


    def texturesReady(self):
        """Returns ``True`` if the ``imageTexture`` and ``clipTexture`` (if
        applicable) are both ready to be used, ``False`` otherwise.
        """
        imageTexReady = (self.imageTexture is not None and
                         self.imageTexture.ready())
        return imageTexReady and self.auxmgr.texturesReady()


    @property
    def clipTexture(self):
        """Returns the :class:`.ImageTexture` associated with the
        :attr:`.VolumeOpts.clipImage`.
        """
        return self.auxmgr.texture('clip')


    @property
    def modulateTexture(self):
        """Returns the :class:`.ImageTexture` associated with the
        :attr:`.VolumeOpts.modulateImage`.
        """
        return self.auxmgr.texture('modulate')


    def addDisplayListeners(self):
        """Called by :meth:`__init__`.

        Adds a bunch of listeners to the :class:`.Display` object, and the
        associated :class:`.VolumeOpts` instance, which define how the image
        should be displayed.

        This is done so we can update the colour, vertex, and image data when
        display properties are changed.
        """

        display = self.display
        opts    = self.opts
        name    = self.name

        crPVs   = opts.getPropVal('clippingRange').getPropertyValueList()

        display .addListener('alpha',            name, self._alphaChanged)
        opts    .addListener('displayRange',     name,
                             self._displayRangeChanged)
        opts    .addListener('modulateRange',    name,
                             self._modulateRangeChanged)

        crPVs[0].addListener(name, self._lowClippingRangeChanged)
        crPVs[1].addListener(name, self._highClippingRangeChanged)

        opts    .addListener('clipImage',        name, self._clipImageChanged)
        opts    .addListener('modulateImage',    name,
                             self._modulateImageChanged)
        opts    .addListener('invertClipping',   name,
                             self._invertClippingChanged)
        opts    .addListener('cmap',             name, self._cmapChanged)
        opts    .addListener('gamma',            name, self._cmapChanged)
        opts    .addListener('logScale',         name, self._cmapChanged)
        opts    .addListener('interpolateCmaps', name, self._cmapChanged)
        opts    .addListener('negativeCmap',     name, self._cmapChanged)
        opts    .addListener('cmapResolution',   name, self._cmapChanged)
        opts    .addListener('useNegativeCmap',  name,
                             self._useNegativeCmapChanged)
        opts    .addListener('invert',           name, self._invertChanged)
        opts    .addListener('modulateAlpha',    name,
                             self._modulateAlphaChanged)
        opts    .addListener('invertModulateAlpha', name,
                             self._modulateAlphaChanged)
        opts    .addListener('volume',           name, self._volumeChanged)
        opts    .addListener('channel',          name, self._channelChanged)
        opts    .addListener('interpolation',    name,
                             self._interpolationChanged)
        opts    .addListener('transform',        name, self._transformChanged)
        opts    .addListener('displayXform',     name,
                             self._displayXformChanged)
        opts    .addListener('enableOverrideDataRange',  name,
                             self._enableOverrideDataRangeChanged)
        opts    .addListener('overrideDataRange', name,
                             self._overrideDataRangeChanged)

        # 3D-only options
        if self.threedee:

            opts.addListener('numSteps',        name, self._numStepsChanged)
            opts.addListener('numInnerSteps',   name,
                             self._numInnerStepsChanged)
            opts.addListener('resolution',      name,  self._resolutionChanged)
            opts.addListener('blendFactor',     name,
                             self._blendPropertiesChanged)
            opts.addListener('blendByIntensity', name,
                             self._blendPropertiesChanged)
            opts.addListener('smoothing',       name, self._smoothingChanged)
            opts.addListener('showClipPlanes',  name,
                             self._showClipPlanesChanged)
            opts.addListener('numClipPlanes',
                             name,
                             self._numClipPlanesChanged)
            opts.addListener('clipMode',        name, self._clipModeChanged)
            opts.addListener('clipPosition',    name, self._clipping3DChanged)
            opts.addListener('clipAzimuth',     name, self._clipping3DChanged)
            opts.addListener('clipInclination', name, self._clipping3DChanged)

        # GLVolume instances need to keep track of whether
        # the volume/channel properties of their corresponding
        # VolumeOpts instance is synced to other VolumeOpts
        # instances - if it is, there an ImageTexture for
        # the image may already exist (i.e. have been
        # created by another GLVolume), and we can just
        # re-use it. Otherwise we will need to create our
        # own ImageTexture.
        #
        # Save a flag so the removeDisplayListeners
        # method knows whether it needs to de-register
        # sync change listeners - using opts.getParent()
        # as the test in that method is dangerous, as
        # the DisplayOpts instance might have already
        # had its destroy method called on it, and might
        # have been detached from its parent.
        self.__syncListenersRegistered = opts.getParent() is not None

        if self.__syncListenersRegistered:
            opts.addSyncChangeListener('volume',
                                       name,
                                       self._imageSyncChanged)
            opts.addSyncChangeListener('channel',
                                       name,
                                       self._imageSyncChanged)


    def removeDisplayListeners(self):
        """Called by :meth:`destroy`. Removes all the property listeners that
        were added by :meth:`addDisplayListeners`.
        """

        display = self.display
        opts    = self.opts
        name    = self.name
        crPVs   = opts.getPropVal('clippingRange').getPropertyValueList()

        display .removeListener('alpha',                   name)
        opts    .removeListener('displayRange',            name)
        opts    .removeListener('modulateRange',           name)
        crPVs[0].removeListener(name)
        crPVs[1].removeListener(name)
        opts    .removeListener('clipImage',               name)
        opts    .removeListener('modulateImage',           name)
        opts    .removeListener('invertClipping',          name)
        opts    .removeListener('cmap',                    name)
        opts    .removeListener('gamma',                   name)
        opts    .removeListener('logScale',                name)
        opts    .removeListener('interpolateCmaps',        name)
        opts    .removeListener('negativeCmap',            name)
        opts    .removeListener('useNegativeCmap',         name)
        opts    .removeListener('cmapResolution',          name)
        opts    .removeListener('invert',                  name)
        opts    .removeListener('modulateAlpha',           name)
        opts    .removeListener('invertModulateAlpha',     name)
        opts    .removeListener('volume',                  name)
        opts    .removeListener('channel',                 name)
        opts    .removeListener('interpolation',           name)
        opts    .removeListener('transform',               name)
        opts    .removeListener('displayXform',            name)
        opts    .removeListener('enableOverrideDataRange', name)
        opts    .removeListener('overrideDataRange',       name)

        if self.threedee:
            opts.removeListener('numSteps',         name)
            opts.removeListener('numInnerSteps',    name)
            opts.removeListener('resolution',       name)
            opts.removeListener('blendFactor',      name)
            opts.removeListener('blendByIntensity', name)
            opts.removeListener('smoothing',        name)
            opts.removeListener('showClipPlanes',   name)
            opts.removeListener('numClipPlanes',    name)
            opts.removeListener('clipMode',         name)
            opts.removeListener('clipPosition',     name)
            opts.removeListener('clipAzimuth',      name)
            opts.removeListener('clipInclination',  name)

        if self.__syncListenersRegistered:
            opts.removeSyncChangeListener('volume',  name)
            opts.removeSyncChangeListener('channel', name)


    def testUnsynced(self):
        """Used by the :meth:`refreshImageTexture` method.

        Returns ``True`` if certain critical :class:`VolumeOpts` properties
        have been unsynced from the parent instance, meaning that this
        ``GLVolume`` instance needs to create its own image texture;
        returns ``False`` otherwise.
        """
        is4D  = len(self.image.shape) >= 4 and self.image.shape[3] > 1
        isRGB = self.image.nvals > 1

        return ((self.opts.getParent() is None)                      or
                (is4D  and not self.opts.isSyncedToParent('volume')) or
                (isRGB and not self.opts.isSyncedToParent('channel')))


    def updateShaderState(self, *args, **kwargs):
        """Calls :func:`.gl14.glvolume_funcs.updateShaderState` or
        :func:`.gl21.glvolume_funcs.updateShaderStatea`, then
        :meth:`.Notifier.notify`. Uses the :func:`.idle.idleWhen` function to
        make sure that it is not called until :meth:`ready` returns ``True``.

        :arg alwaysNotify: Must be passed as a keyword argument. If
                           ``False`` (the default), ``notify`` is only called
                           if ``glvolume_funcs.updateShaderState`` returns
                           ``True``. Otherwise, ``notify`` is always called.
        """

        alwaysNotify = kwargs.pop('alwaysNotify', None)

        # When alwaysNotify is True, we
        # set a flag on this GLVolume
        # instance to make sure that the
        # func() function below (which is
        # called asynchronously) gets
        # its value.
        #
        # We have to do this because this
        # updateShaderState method may be
        # called multiple times for a single
        # event, with different values of
        # alwaysNotify, and some of these
        # calls may be silently dropped
        # (see below).
        #
        # But if one of those calls needs
        # to force a notification, we want
        # that notification to happen.
        if alwaysNotify:
            self.__alwaysNotify = True

        def func():
            if fslgl.glvolume_funcs.updateShaderState(self) or \
               self.__alwaysNotify:
                self.notify()
                self.__alwaysNotify = False

        # Don't re-queue the update if it is
        # already queued on the idle loop.
        # As mentioned above, updateShaderState
        # may get called several times for a
        # single event, but in this situation
        # we only want to actually do the
        # update once.
        idle.idleWhen(func,
                      self.ready,
                      name=self.name,
                      skipIfQueued=True)


    def refreshImageTexture(self, **kwargs):
        """Refreshes the :class:`.ImageTexture` used to store the
        :class:`.Image` data. This is performed through the :mod:`.resources`
        module, so the image texture can be shared between multiple
        ``GLVolume`` instances.

        All keyword arguments are passed through to the :class:`.ImageTexture`
        constructor.
        """

        opts     = self.opts
        texName  = self.texName
        unsynced = self.testUnsynced()

        if unsynced:
            texName = '{}_unsync_{}'.format(texName, id(opts))

        if self.imageTexture is not None:

            if self.imageTexture.name == texName:
                return None

            self.imageTexture.deregister(self.name)
            glresources.delete(self.imageTexture.name)

        if opts.interpolation == 'none': interp = gl.GL_NEAREST
        else:                            interp = gl.GL_LINEAR

        if opts.enableOverrideDataRange: normRange = opts.overrideDataRange
        else:                            normRange = None

        self.imageTexture = glresources.get(
            texName,
            textures.createImageTexture,
            texName,
            self.image,
            interp=interp,
            channel=opts.channel,
            volume=opts.index()[3:],
            normaliseRange=normRange,
            notify=False,
            **kwargs)

        self.imageTexture.register(self.name, self.__texturesChanged)


    def registerAuxImage(self, which, image, onReady=None):
        """Calls :meth:`.AuxImageTextureManager.registerAuxImage`, making
        sure that the texture interpolation is set appropriately.
        """

        if self.opts.interpolation == 'none': interp = gl.GL_NEAREST
        else:                                 interp = gl.GL_LINEAR

        self.auxmgr.texture(which).deregister(self.name)
        self.auxmgr.registerAuxImage(which, image, interp=interp)
        self.auxmgr.texture(which).register(self.name, self.__texturesChanged)
        if onReady is not None:
            idle.idleWhen(onReady, self.auxmgr.texturesReady)


    def refreshColourTextures(self):
        """Refreshes the :class:`.ColourMapTexture` instances used to colour
        image voxels.
        """

        display  = self.display
        opts     = self.opts
        alpha    = display.alpha / 100.0
        cmap     = opts.cmap
        interp   = opts.interpolateCmaps
        res      = opts.cmapResolution
        logScale = opts.logScale
        gamma    = opts.realGamma(opts.gamma)
        negCmap  = opts.negativeCmap
        invert   = opts.invert
        dmin     = opts.displayRange[0]
        dmax     = opts.displayRange[1]

        if interp: interp = gl.GL_LINEAR
        else:      interp = gl.GL_NEAREST

        self.colourTexture.set(cmap=cmap,
                               invert=invert,
                               alpha=alpha,
                               resolution=res,
                               gamma=gamma,
                               logScale=logScale,
                               interp=interp,
                               displayRange=(dmin, dmax))

        self.negColourTexture.set(cmap=negCmap,
                                  invert=invert,
                                  alpha=alpha,
                                  resolution=res,
                                  gamma=gamma,
                                  interp=interp,
                                  displayRange=(dmin, dmax))


    def preDraw(self):
        """Binds the :class:`.ImageTexture` to ``GL_TEXTURE0`` and the
        :class:`.ColourMapTexture` to ``GL_TEXTURE1, and calls the
        version-dependent ``preDraw`` function.
        """

        # Set up the image and colour textures
        self.imageTexture    .bindTexture(gl.GL_TEXTURE0)
        self.colourTexture   .bindTexture(gl.GL_TEXTURE1)
        self.negColourTexture.bindTexture(gl.GL_TEXTURE2)
        self.clipTexture     .bindTexture(gl.GL_TEXTURE3)
        self.modulateTexture .bindTexture(gl.GL_TEXTURE4)

        fslgl.glvolume_funcs.preDraw(self)


    def draw2D(self, canvas, *args, **kwargs):
        """Calls the version dependent ``draw2D`` function. """

        with glroutines.enabled((gl.GL_CULL_FACE)):
            gl.glPolygonMode(gl.GL_FRONT_AND_BACK, gl.GL_FILL)
            gl.glCullFace(gl.GL_BACK)
            gl.glFrontFace(self.frontFace(canvas))
            fslgl.glvolume_funcs.draw2D(self, canvas, *args, **kwargs)


    def draw3D(self, canvas, xform=None):
        """Calls the version dependent ``draw3D`` function. """

        opts = self.opts
        w, h = canvas.GetScaledSize()
        res  = self.opts.resolution / 100.0
        sw   = int(np.ceil(w * res))
        sh   = int(np.ceil(h * res))

        # Initialise and resize
        # the offscreen textures
        for rt in [self.renderTexture1, self.renderTexture2]:
            rt.shape = sw, sh
            with rt.target():
                glroutines.clear((0, 0, 0, 0))

        if self.opts.resolution != 100:
            gl.glViewport(0, 0, sw, sh)

        # Do the render. Even though we're
        # drawing off-screen,  we need to
        # enable depth-testing, otherwise
        # depth values will not get written
        # to the depth buffer!
        #
        # The glvolume_funcs.draw3D function
        # will put the final render into
        # renderTexture1
        with glroutines.enabled((gl.GL_DEPTH_TEST, gl.GL_CULL_FACE)):
            gl.glPolygonMode(gl.GL_FRONT_AND_BACK, gl.GL_FILL)
            gl.glFrontFace(gl.GL_CCW)
            gl.glCullFace(gl.GL_BACK)
            fslgl.glvolume_funcs.draw3D(self, canvas, xform)

        if self.opts.resolution != 100:
            gl.glViewport(0, 0, w, h)

        # Apply smoothing if needed. If smoothing
        # is enabled, the final render will be in
        # renderTexture2
        if opts.smoothing > 0:
            self.smoothFilter.set(offsets=[1.0 / sw, 1.0 / sh])
            self.smoothFilter.osApply(self.renderTexture1,
                                      self.renderTexture2)

        # We now have the final result
        # - draw it to the screen.
        verts = np.array([[-1, -1, 0],
                          [-1,  1, 0],
                          [ 1, -1, 0],
                          [ 1, -1, 0],
                          [-1,  1, 0],
                          [ 1,  1, 0]], dtype=np.float32)

        with glroutines.enabled(gl.GL_DEPTH_TEST):

            # If smoothing was not applied, rt1
            # contains the final render. Otherwise,
            # rt2 contains the final render, but rt1
            # contains the depth information. So we
            # need to temporarily replace rt2.depth
            # with rt1.depth.
            if opts.smoothing > 0:
                src    = self.renderTexture2
                olddep = self.renderTexture2.depthTexture
                dep    = self.renderTexture1.depthTexture
            else:
                src    = self.renderTexture1
                olddep = self.renderTexture1.depthTexture
                dep    = olddep

            src.depthTexture = dep
            src.draw(verts, useDepth=True)
            src.depthTexture = olddep


    def drawAll(self, *args, **kwargs):
        """Calls the version dependent ``drawAll`` function. """
        fslgl.glvolume_funcs.drawAll(self, *args, **kwargs)


    def postDraw(self):
        """Unbinds the ``ImageTexture`` and ``ColourMapTexture``, and calls the
        version-dependent ``postDraw`` function.
        """

        self.imageTexture    .unbindTexture()
        self.colourTexture   .unbindTexture()
        self.negColourTexture.unbindTexture()
        self.clipTexture     .unbindTexture()
        self.modulateTexture .unbindTexture()

        fslgl.glvolume_funcs.postDraw(self)


    def getAuxTextureXform(self, which):
        """Calculates a transformation matrix which will transform from the
        image coordinate system into the :attr:`.VolumeOpts.clipImage` or
        :attr:`.VolumeOpts.modulateImage` coordinate system. If the property
        is ``None``, it will be an identity transform.

        This transform is used by shader programs to find the auxillary image
        coordinates that correspond with specific image coordinates.
        """
        return affine.concat(
            self.auxmgr.textureXform(which),
            # to support 2D image textures
            self.imageTexture.invTexCoordXform(self.overlay.shape))


    def getModulateValueXform(self):
        """Returns an affine transform to normalise alpha modulation values.

        The GL volume shaders need to normalise the modulate value by the
        modulation range to generate an opacity value. We calculate a suitable
        scale and offset by buildin an affine transform which transforms voxel
        values from the image/modulate image texture range to 0/1, where 0
        corresponds to the low modulate range bound, and 1 to the high
        modulate range bound. The resulting scale/offset can be used by the
        shader to convert a modulate value directly into an opacity value.
        """

        opts = self.opts
        if opts.modulateImage is None:
            modXform = self.imageTexture.voxValXform
        else:
            modXform = self.modulateTexture.voxValXform

        modXform = affine.concat(
            opts.modulateScaleOffset(xform=True),
            modXform)

        return modXform


    def generateVertices2D(self, zpos, axes, bbox=None):
        """Overrides :meth:`.GLImageObject.generateVertices2D`.

        Appliies the :meth:`.ImageTextureBase.texCoordXform` to the texture
        coordinates - this is performed to support 2D images/textures.
        """

        vertices, voxCoords, texCoords = \
            glimageobject.GLImageObject.generateVertices2D(
                self, zpos, axes, bbox)

        texCoords = affine.transform(
            texCoords, self.imageTexture.texCoordXform(self.overlay.shape))

        return vertices, voxCoords, texCoords


    def generateVertices3D(self, bbox=None):
        """Overrides :meth:`.GLImageObject.generateVertices3D`.

        Appliies the :meth:`.ImageTextureBase.texCoordXform` to the texture
        coordinates - this is performed to support 2D images/textures.
        """

        vertices, voxCoords, texCoords = \
            glimageobject.GLImageObject.generateVertices3D(self, bbox)

        texCoords = affine.transform(
            texCoords, self.imageTexture.texCoordXform(self.overlay.shape))

        return vertices, voxCoords, texCoords


    def _alphaChanged(self, *a):
        """Called when the :attr:`.Display.alpha` property changes. """

        self.refreshColourTextures()
        if self.threedee:
            self.updateShaderState(alwaysNotify=True)
        else:
            self.notify()


    def _displayRangeChanged(self, *a):
        """Called when the :attr:`.VolumeOpts.displayRange` property changes.
        """
        self.refreshColourTextures()
        self.updateShaderState()


    def _lowClippingRangeChanged(self, *a):
        """Called when the low :attr:`.VolumeOpts.clippingRange` property
        changes. Separate listeners are used for the low and high clipping
        values to avoid unnecessary duplicate refreshes in the event that the
        :attr:`.VolumeOpts.linkLowRanges` or
        :attr:`.VolumeOpts.linkHighRanges` flags are set.
        """
        if self.opts.linkLowRanges:
            return

        self.updateShaderState()


    def _highClippingRangeChanged(self, *a):
        """Called when the high :attr:`.VolumeOpts.clippingRange` property
        changes (see :meth:`_lowClippingRangeChanged`).
        """
        if self.opts.linkHighRanges:
            return

        self.updateShaderState(self)


    def _modulateRangeChanged(self, *a):
        """Called when the :attr:`.VolumeOpts.modulateRange` property changes.
        """
        self.updateShaderState()


    def _clipImageChanged(self, *a):
        """Called when the :attr:`.VolumeOpts.clipImage` property changes.
        """
        self.registerAuxImage('clip', self.opts.clipImage)


    def _modulateImageChanged(self, *a):
        """Called when the :attr:`.VolumeOpts.modulateImage` property changes.
        """
        self.registerAuxImage('modulate', self.opts.modulateImage)


    def _invertClippingChanged(self, *a):
        """Called when the :attr:`.VolumeOpts.invertClipping` property changes.
        """
        self.updateShaderState()


    def _cmapChanged(self, *a):
        """Called when the :attr:`.VolumeOpts.cmap` or
        :attr:`.VolumeOpts.negativeCmap` properties change.
        """
        self.refreshColourTextures()
        self.notify()


    def _useNegativeCmapChanged(self, *a):
        """Called when the :attr:`.VolumeOpts.useNegativeCmap` property
        changes.
        """
        self.updateShaderState()


    def _invertChanged(self, *a):
        """Called when the :attr:`.VolumeOpts.invert` property changes. """
        self.refreshColourTextures()
        self.notify()


    def _modulateAlphaChanged(self, *a):
        """Called when the :attr:`.VolumeOpts.modulateAlpha` or
        :attr:`.VolumeOpts.invertModulateAlpha` properties change.
        Calls :meth:`updateShaderState`.
        """
        self.updateShaderState()


    def _enableOverrideDataRangeChanged(self, *a):
        """Called when the :attr:`.VolumeOpts.enableOverrideDataRange` property
        changes. Calls :meth:`_volumeChanged`.
        """
        self._volumeChanged(volRefresh=True)


    def _overrideDataRangeChanged(self, *a):
        """Called when the :attr:`.VolumeOpts.overrideDataRange` property
        changes. Calls :meth:`_volumeChanged`, but only if
        :attr:`.VolumeOpts.enableOverrideDataRange` is ``True``.
        """
        if self.opts.enableOverrideDataRange:
            self._volumeChanged(volRefresh=True)


    def _volumeChanged(self, *a, **kwa):
        """Called when the :attr:`.NiftiOpts.volume` property changes Also
        called when other properties, which require a texture refresh, change.
        """
        opts       = self.opts
        volRefresh = kwa.pop('volRefresh', False)

        if opts.interpolation == 'none': interp = gl.GL_NEAREST
        else:                            interp = gl.GL_LINEAR

        if opts.enableOverrideDataRange: normRange = opts.overrideDataRange
        else:                            normRange = None

        self.imageTexture.set(volume=opts.index()[3:],
                              channel=opts.channel,
                              interp=interp,
                              volRefresh=volRefresh,
                              normaliseRange=normRange)

        self.clipTexture    .set(interp=interp)
        self.modulateTexture.set(interp=interp)


    def _channelChanged(self, *a, **kwa):
        """Called when the :attr:`.NiftiOpts.channel` changes.
        Refreshes the texture.
        """
        self._volumeChanged()


    def _interpolationChanged(self, *a):
        """Called when the :attr:`.NiftiOpts.interpolation` property changes.
        """
        self._volumeChanged(volRefresh=True)


    def _transformChanged(self, *a):
        """Called when the :attr:`.NiftiOpts.transform` property changes.
        """
        self.notify()


    def _displayXformChanged(self, *a):
        """Called when the :attr:`.NiftiOpts.displayXform` property changes.
        """
        self.notify()


    def _numStepsChanged(self, *a):
        """Called when the :attr:`.Volume3DOpts.numSteps` property changes.
        """
        self.notify()


    def _numInnerStepsChanged(self, *a):
        """Called when the :attr:`.Volume3DOpts.numInnerSteps` property
        changes.
        """
        fslgl.glvolume_funcs.compileShaders(self)
        self.updateShaderState(alwaysNotify=True)


    def _resolutionChanged(self, *a):
        """Called when the :attr:`.Volume3DOpts.resolution` property
        changes.
        """
        self.notify()


    def _numClipPlanesChanged(self, *a):
        """Called when the :attr:`.Volume3DOpts.numClipPlanes` property
        changes.
        """
        if float(fslgl.GL_COMPATIBILITY) == 1.4:
            fslgl.glvolume_funcs.compileShaders(self)
        self.updateShaderState(alwaysNotify=True)


    def _clipModeChanged(self, *a):
        """Called when the :attr:`.Volume3DOpts.clipMode` property
        changes.
        """
        if float(fslgl.GL_COMPATIBILITY) == 1.4:
            fslgl.glvolume_funcs.compileShaders(self)
        self.updateShaderState(alwaysNotify=True)


    def _clipping3DChanged(self, *a):
        """Called when any of the :attr:`.Volume3DOpts.clipPosition`,
        :attr:`.Volume3DOpts.clipAzimuth`, or
        :attr:`.Volume3DOpts.clipInclination` properties change.
        """

        self.updateShaderState(alwaysNotify=True)


    def _showClipPlanesChanged(self, *a):
        """Called when the :attr:`.Volume3DOpts.showClipPlanes` property
        changes.
        """
        self.updateShaderState(alwaysNotify=True)


    def _blendPropertiesChanged(self, *a):
        """Called when the :attr:`.Volume3DOpts.showClipPlanes` property
        changes.
        """
        self.updateShaderState(alwaysNotify=True)


    def _smoothingChanged(self, *a):
        """Called when the :attr:`.Volume3DOpts.smoothing` property
        changes.
        """
        self.smoothFilter.set(kernSize=self.opts.smoothing * 2)
        self.notify()


    def _imageSyncChanged(self, *a):
        """Called when the synchronisation state of the
        :attr:`.NiftiOpts.volume` or :attr:`.VolumeOpts.interpolation`
        properties change.
        """

        self.refreshImageTexture()
        self.updateShaderState(alwaysNotify=True)


    def __texturesChanged(self, *a):
        """Called when either the ``imageTexture`` or the ``clipTexture``
        changes. Calls :meth:`updateShaderState`.
        """
        self.updateShaderState(alwaysNotify=True)



#!/usr/bin/env python
#
# scene3dcanvas.py - The Scene3DCanvas class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`.Scene3DCanvas` class, which is used by
FSLeyes for its 3D view.

On each refresh the ``Scene3DCanvas`` draws each
overlay to an offscreen 2D texture.
For OpenGL 2.1 and newer, all of these textures are then blended together
according to their depth using a custom shader program.

For OpenGL 1.4, the textures are then just drawn directly to the screen
according to the current overlay order.
"""


import logging

import numpy     as np
import OpenGL.GL as gl

import fsl.data.mesh        as fslmesh
import fsl.data.image       as fslimage
import fsl.utils.idle       as idle
import fsl.transform.affine as affine

import fsleyes.data.tractogram           as fsltractogram
import fsleyes.gl                        as fslgl
import fsleyes.gl.routines               as glroutines
import fsleyes.gl.globject               as globject
import fsleyes.gl.annotations            as annotations
import fsleyes.gl.shaders                as shaders
import fsleyes.gl.text                   as gltext
import fsleyes.gl.textures               as textures
import fsleyes.displaycontext            as fsldisplay
import fsleyes.displaycontext.canvasopts as canvasopts


log = logging.getLogger(__name__)


class Scene3DCanvas:
    """The ``Scene3DCanvas`` is an OpenGL canvas used to draw overlays in a 3D
    view. Currently only ``volume``, ``mesh``, and ``tractogram`` overlay
    types are supported.
    """

    def __init__(self, overlayList, displayCtx):

        self.__name           = '{}_{}'.format(type(self).__name__, id(self))
        self.__opts           = canvasopts.Scene3DCanvasOpts()
        self.__overlayList    = overlayList
        self.__displayCtx     = displayCtx
        self.__annotations    = annotations.Annotations(self)
        self.__viewMat        = np.eye(4)
        self.__projMat        = np.eye(4)
        self.__invViewProjMat = np.eye(4)
        self.__viewport       = None
        self.__resetLightPos  = True

        # Dictionary of GLObjects and off-screen
        # RenderTextures for every overlay
        self.__glObjects      = {}
        self.__globjTextures  = {}

        # Shader program used to blend overlays,
        # only used for OpenGL >= 2.1. The number
        # of overlays that the shader program can
        # blend is hard-coded, so we maintain a
        # a set of them, one for each possible
        # number of overlays.
        #
        # n.b. It should be possible to do this
        # this for GL14, but would require a
        # sorting algorithm to be implemented
        # in ARB assembly, so has been left as
        # a possible future task.
        self.__shaders = {}
        self.__compileShaders()

        # gl.text.Text objects containing anatomical
        # orientation labels ordered
        # (xlo, xhi, ylo, yhi, zlo, zhi) where xyz
        # are the display coordinate system axes.
        # Created in __refreshLegendLabels
        self.__legendLabels = None

        overlayList.addListener('overlays',
                                self.__name,
                                self.__overlayListChanged)
        displayCtx.addListener('bounds',
                               self.__name,
                               self.__displayBoundsChanged)

        opts = self.opts
        opts.addListener('pos',           self.__name, self.Refresh)
        opts.addListener('showCursor',    self.__name, self.Refresh)
        opts.addListener('cursorColour',  self.__name, self.Refresh)
        opts.addListener('bgColour',      self.__name, self.Refresh)
        opts.addListener('showLegend',    self.__name, self.Refresh)
        opts.addListener('legendColour',  self.__name,
                         self.__refreshLegendLabels)
        opts.addListener('labelSize',     self.__name,
                         self.__refreshLegendLabels)
        opts.addListener('zoom',          self.__name, self.Refresh)
        opts.addListener('offset',        self.__name, self.Refresh)
        opts.addListener('rotation',      self.__name, self.Refresh)
        opts.addListener('showLight',     self.__name, self.Refresh)
        opts.addListener('light',         self.__name, self.Refresh)
        opts.addListener('lightPos',      self.__name, self.Refresh)
        opts.addListener('lightDistance', self.__name, self.Refresh)
        opts.addListener('highDpi',       self.__name, self.__highDpiChanged)


    def __compileShaders(self):
        """(Re-)compiles the shader programs which are used to blend overlays
        together. The shader programs are only used for OpenGL 2.1 and newer.

        This method is called every time the overlay list changes as the
        number of overlays is hard-coded into the shader programs.
        """

        if float(fslgl.GL_COMPATIBILITY) < 2.1:
            return

        noverlays = len(self.__overlayList)
        if noverlays == 0:
            return

        vertSrc = shaders.getVertexShader(  'scene3dcanvas')
        fragSrc = shaders.getFragmentShader('scene3dcanvas')

        # We maintain separate shader prorgams
        # for any possible number of overlays
        # (as not all loaded overlays may be
        # displayed)
        for n in range(1, noverlays + 1):
            if n in self.__shaders:
                continue

            env    = {'noverlays' : n}
            shader = shaders.GLSLShader(vertSrc, fragSrc, constants=env)
            self.__shaders[n] = shader

            with shader.loaded():
                for i in range(n):
                    shader.set(f'rgba_{i}',  i * 2)
                    shader.set(f'depth_{i}', i * 2 + 1)


    def destroy(self):
        """Must be called when this Scene3DCanvas is no longer used. """

        if self.destroyed:
            return

        self.__overlayList.removeListener('overlays', self.__name)
        self.__displayCtx .removeListener('bounds',   self.__name)

        for ovl in list(self.__glObjects.keys()):
            self.__deregisterOverlay(ovl)

        if self.__legendLabels is not None:
            for lbl in self.__legendLabels:
                lbl.destroy()

        self.__annotations.destroy()

        for shader in self.__shaders.values():
            shader.destroy()

        self.__annotations  = None
        self.__shaders      = None
        self.__opts         = None
        self.__displayCtx   = None
        self.__overlayList  = None
        self.__glObjects    = None
        self.__legendLabels = None


    @property
    def destroyed(self):
        """Returns ``True`` if :meth:`destroy` has been called. """
        return self.__overlayList is None


    @property
    def opts(self):
        """Returns a reference to the :class:`.Scene3DCanvasOpts` instance.
        """
        return self.__opts


    @property
    def displayCtx(self):
        """Returns a reference to the :class:`.DisplayContext` for this canvas.
        """
        return self.__displayCtx


    def getAnnotations(self):
        """Return the :class:`.Annotations` object, for drawing annotations
        on this ``Scene3DCanvas``.
        """
        return self.__annotations


    @property
    def lightPos(self):
        """Takes the values of :attr:`.Scene3DOpts.lightPos` and
        :attr:`.Scene3DOpts.lightDistance`, and converts it to a position in
        the display coordinate system. The ``Scene3DOpts.lightPos`` property
        contains rotations about the centre of the display bounding box,
        and the :attr:`.Scene3DOpts.lightDistance` property specifies the
        distance of the light from the bounding box centre.
        """
        b        = self.__displayCtx.bounds
        centre   = np.array([b.xlo + 0.5 * (b.xhi - b.xlo),
                             b.ylo + 0.5 * (b.yhi - b.ylo),
                             b.zlo + 0.5 * (b.zhi - b.zlo)])

        yaw, pitch, roll = self.opts.lightPos
        distance         = self.opts.lightDistance
        yaw              = yaw   * np.pi / 180
        pitch            = pitch * np.pi / 180
        roll             = roll  * np.pi / 180

        rotmat = affine.axisAnglesToRotMat(pitch, roll, yaw)
        xform  = affine.compose([1, 1, 1],
                                [0, 0, 0],
                                rotmat,
                                origin=centre)

        lightPos = centre + [0, 0, distance * b.zlen]
        lightPos = affine.transform(lightPos, xform)

        return lightPos


    @property
    def resetLightPos(self):
        """By default, the :attr:`lightPos` is updated whenever the
        :attr:`.DisplayContext.bounds` change. This flag can be used to
        disable this behaviour.
        """
        return self.__resetLightPos


    @resetLightPos.setter
    def resetLightPos(self, reset):
        """Control whether the :attr:`lightPos` property is reset whenever
        the :attr:`.DisplayContext.bounds` change.
        """
        self.__resetLightPos = reset


    def defaultLightPos(self):
        """Resets the :attr:`lightPos` property to a sensible value. """
        self.opts.lightPos = [0, 0, 0]


    @property
    def viewMatrix(self):
        """Returns the view matrix for the current scene - this is an affine
        matrix which encodes the current :attr:`.Scene3DCanvasOpts.offset`,
        :attr:`.Scene3DCanvasOpts.zoom`,
        :attr:`.Scene3DCanvasOpts.rotation` and camera settings.

        See :meth:`__genViewMatrix`.
        """
        return self.__viewMat


    @property
    def viewScale(self):
        """Returns an affine matrix which encodes the current
        :attr:`.Scene3DCanvasOpts.zoom` setting.
        """
        return self.__viewScale


    @property
    def viewOffset(self):
        """Returns an affine matrix which encodes the current
        :attr:`.Scene3DCanvasOpts.offset` setting.
        """
        return self.__viewOffset


    @property
    def viewRotation(self):
        """Returns an affine matrix which encodes the current
        :attr:`.Scene3DCanvasOpts.rotation` setting.
        """
        return self.__viewRotate


    @property
    def viewCamera(self):
        """Returns an affine matrix which encodes the current camera.
        transformation. The initial camera orientation in the view shown by a
        :class:`Scene3DCanvas` is located on the positive Y axis, is oriented
        towards the positive Z axis, and is pointing towards the centre of the
        :attr:`.DisplayContext.displayBounds`.
        """
        return self.__viewCamera


    @property
    def projectionMatrix(self):
        """Returns the projection matrix. This is an affine matrix which
        converts from normalised device coordinates (NDCs, coordinates between
        -1 and +1) into viewport coordinates. The initial viewport for a
        :class:`Scene3DCanvas` is configured by the :func:`.routines.ortho3D`
        function.

        See :meth:`__setViewport`.
        """
        return self.__projMat


    @property
    def mvpMatrix(self):
        """Returns the current model*view*projection matrix. """
        return affine.concat(self.__projMat, self.__viewMat)


    @property
    def invViewProjectionMatrix(self):
        """Returns the inverse of the model-view-projection matrix, the
        equivalent of:

        ``invert(projectionMatrix * viewMatrix)``
        """
        return self.__invViewProjMat


    @property
    def viewport(self):
        """Returns a list of three ``(min, max)`` tuples which specify the
        viewport limits of the currently displayed scene.
        """
        return self.__viewport


    def pixelSize(self):
        """Returns the (approximate) size (as a tuple containing
        ``(width, height)`` of one pixel in the display coordinate system.
        """

        # The right thing to do would be to transform
        # a vector by the inverse mvp matrix. But just
        # using the initial X/Y bounds (before
        # rotatoin/scaling), then applying the zoom/
        # scaling factor should work.
        w, h     = self.GetSize()
        zoom     = self.viewScale[0, 0]
        xlo, xhi = self.viewport[0]
        ylo, yhi = self.viewport[1]

        pw = (xhi - xlo) / zoom / w
        ph = (yhi - ylo) / zoom / h

        return pw, ph


    def canvasToWorld(self, xpos, ypos, near=True):
        """Transform the given x/y canvas coordinates into the display
        coordinate system. The calculated coordinates will be located on
        the near clipping plane.

        :arg near: If ``True`` (the default), the returned coordinate will
                   be located on the near clipping plane. Otherwise, the
                   coordinate will be located on the far clipping plane.
        """

        width, height = self.GetSize()

        # Normalise pixels to [-1, 1]
        xp = -1 + 2.0 * xpos / width
        yp = -1 + 2.0 * ypos / height

        # We set the Z coord so the resulting
        # coordinates will be located on either
        # the  near or clipping planes.
        if near: pos = [xp, yp, -1]
        else:    pos = [xp, yp,  1]

        # The first step is to convert mouse
        # coordinates from [-1, 1] to viewport
        # coodinates via the inverse projection
        # matrix.

        # The second step is to transform from
        # viewport coords into model-view coords.
        # This is easy - transform by the inverse
        # MV matrix.

        # We perform both of these steps in one
        # by concatenating then inverting the
        # view/projection matrices. This is
        # calculated and cached for us in the
        # __setViewport method.
        pos = affine.transform(pos, self.__invViewProjMat)

        return pos


    def getGLObject(self, overlay):
        """Returns the :class:`.GLObject` associated with the given overlay,
        or ``None`` if there is not one.
        """
        return self.__glObjects.get(overlay, None)


    def getGLObjects(self):
        """Returns two lists:

         - A list of overlays to be drawn
         - A list of corresponding :class:`GLObject` instances

        This method also creates ``GLObject`` instances for any overlays
        in the :class:`.OverlayList` that do not have one.
        """

        # Sort the overlays so that volumes come last.
        # This will normally makes no difference (when
        # using OpenGL >= 2.1), but is important in
        # OpenGL 1.4 fallback mode to get somewhat
        # sensible blending.
        overlays = self.__displayCtx.getOrderedOverlays()
        vols     = [o for o in overlays if isinstance(o, fslimage.Image)]
        others   = [o for o in overlays if o not in vols]

        overlays = []
        globjs   = []

        for ovl in others + vols:
            globj = self.getGLObject(ovl)

            # If there is no GLObject for this
            # overlay, create one, but don't
            # add it to the list (as creation
            # is done asynchronously).
            if globj is None:
                self.__registerOverlay(ovl)

            # Otherwise, if the value for this
            # overlay evaluates to False, that
            # means that it has been scheduled
            # for creation, but is not ready
            # yet.
            elif globj:
                overlays.append(ovl)
                globjs  .append(globj)

        return overlays, globjs


    def _initGL(self):
        """Called when the canvas is ready to be drawn on. """
        self.__overlayListChanged()
        self.__displayBoundsChanged()


    def __overlayListChanged(self, *a):
        """Called when the :class:`.OverlayList` changes. Destroys/creates
        :class:`.GLObject` instances as necessary.
        """

        # Destroy any GL objects for overlays
        # which are no longer in the list
        for ovl, globj in list(self.__glObjects.items()):
            if ovl not in self.__overlayList:
                self.__deregisterOverlay(ovl)

        # Create GLObjects for any
        # newly added overlays
        for ovl in self.__overlayList:
            if ovl not in self.__glObjects:
                self.__registerOverlay(ovl)

        self.__compileShaders()
        self.__refreshLegendLabels(refresh=False)


    def __refreshLegendLabels(self, *a, refresh=True):
        """Called when the legend labels (anatomical orientations) need
        to be refreshed - when the selected overlay changes, or when
        the :attr:`.Scene3DCanvasOpts.legendColour` is changed.
        """

        # Update legend labels. Figure out the
        # anatomical labels for each axis.
        overlay = self.__displayCtx.getSelectedOverlay()

        if overlay is None:
            return

        if self.__legendLabels is None:
            self.__legendLabels = [gltext.Text(coordinates='pixels')
                                   for _ in range(6)]

        dopts  = self.__displayCtx.getOpts(overlay)
        labels = dopts.getLabels()[0]

        # getLabels returns (xlo, ylo, zlo, xhi, yhi, zhi) -
        # - rearrange them to (xlo, xhi, ylo, yhi, zlo, zhi)
        labels = [labels[0],
                  labels[3],
                  labels[1],
                  labels[4],
                  labels[2],
                  labels[5]]

        for label, text in zip(labels, self.__legendLabels):
            text.text     = label
            text.colour   = self.opts.legendColour
            text.fontSize = self.opts.labelSize
            text.halign   = 'centre'
            text.valign   = 'centre'

        if refresh:
            self.Refresh()


    def __highDpiChanged(self, *a):
        """Called when the :attr:`.Scene3DCanvasOpts.highDpi` property
        changes. Calls the :meth:`.GLCanvasTarget.EnableHighDPI` method.
        """
        self.EnableHighDPI(self.opts.highDpi)


    def __displayBoundsChanged(self, *a):
        """Called when the :attr:`.DisplayContext.bounds` change. Resets
        the :attr:`.Scene3DCanvasOpts.lightPos` property.
        """

        if self.resetLightPos:
            self.defaultLightPos()

        self.Refresh()


    def __registerOverlay(self, overlay):
        """Called when a new overlay has been added to the overlay list.
        Creates a ``GLObject`` for it, and registers property listeners.
        """

        if not isinstance(overlay, (fslmesh.Mesh,
                                    fslimage.Image,
                                    fsltractogram.Tractogram)):
            return

        log.debug('Registering overlay {}'.format(overlay))

        display = self.__displayCtx.getDisplay(overlay)

        if not self.__genGLObject(overlay):
            return

        display.addListener('enabled', self.__name, self.Refresh)
        display.addListener('overlayType',
                            self.__name,
                            self.__overlayTypeChanged)


    def __deregisterOverlay(self, overlay):
        """Called when an overlay has been removed from the overlay list.
        Dstroys the ``GLObject`` for it, and de-registers property listeners.
        """

        log.debug('Deregistering overlay {}'.format(overlay))

        try:
            display = self.__displayCtx.getDisplay(overlay)
            display.removeListener('overlayType', self.__name)
            display.removeListener('enabled',     self.__name)
        except fsldisplay.InvalidOverlayError:
            pass

        globj = self.__glObjects    .pop(overlay, None)
        tex   = self.__globjTextures.pop(overlay, None)

        if globj is not None:
            globj.deregister(self.__name)
            globj.destroy()
        if tex is not None:
            tex.destroy()


    def __genGLObject(self, overlay):
        """Create a ``GLObject`` for the given overlay, if one doesn't already
        exist.
        """

        if overlay in self.__glObjects:
            return False

        display = self.__displayCtx.getDisplay(overlay)

        if display.overlayType not in ('volume', 'mesh', 'tractogram'):
            return False

        self.__glObjects[overlay] = False

        def create():

            if not self or self.destroyed:
                return

            if overlay not in self.__glObjects:
                return

            if not self._setGLContext():
                self.__glObjects.pop(overlay)
                return

            log.debug('Creating GLObject for %s', overlay)

            globj = globject.createGLObject(overlay,
                                            self.__overlayList,
                                            self.__displayCtx,
                                            True)

            if globj is not None:
                globj.register(self.__name, self.Refresh)
                tex = textures.RenderTexture(globj.name, 'cd')
                self.__globjTextures[overlay] = tex
                self.__glObjects[    overlay] = globj

        idle.idle(create)
        return True


    def __overlayTypeChanged(self, value, valid, display, name):
        """Called when the :attr:`.Display.overlayType` of an overlay
        has changed. Re-generates a :class:`.GLObject` for it.
        """

        overlay = display.overlay
        globj   = self.__glObjects    .pop(overlay, None)
        tex     = self.__globjTextures.pop(overlay, None)

        if globj is not None:
            globj.deregister(self.__name)
            globj.destroy()
        if tex is not None:
            tex.destroy()

        self.__genGLObject(overlay)
        self.Refresh()


    def __genViewMatrix(self, w, h):
        """Generate and return a transformation matrix to be used as the
        model-view matrix. This includes applying the current :attr:`zoom`,
        :attr:`rotation` and :attr:`offset` settings, and configuring
        the camera. This method is called by :meth:`__setViewport`.

        :arg w: Canvas width in pixels
        :arg h: Canvas height in pixels
        """

        opts   = self.opts
        b      = self.__displayCtx.bounds
        centre = [b.xlo + 0.5 * b.xlen,
                  b.ylo + 0.5 * b.ylen,
                  b.zlo + 0.5 * b.zlen]

        # The MV matrix comprises (in this order):
        #
        #    - A rotation (the rotation property)
        #
        #    - Camera configuration. With no rotation, the
        #      camera will be looking towards the positive
        #      Y axis (i.e. +y is forwards), and oriented
        #      towards the positive Z axis (i.e. +z is up)
        #
        #    - A translation (the offset property)
        #    - A scaling (the zoom property)

        # Scaling and rotation matrices. Rotation
        # is always around the centre of the
        # displaycontext bounds (the bounding
        # box which contains all loaded overlays).
        scale  = opts.zoom / 100.0
        scale  = affine.scaleOffsetXform([scale] * 3, 0)
        rotate = affine.rotMatToAffine(opts.rotation, centre)

        # The offset property is defined in x/y
        # pixels, normalised to [-1, 1]. We need
        # to convert them into viewport space,
        # where the horizontal axis maps to
        # (-xhalf, xhalf), and the vertical axis
        # maps to (-yhalf, yhalf). See
        # gl.routines.ortho3D.
        offset     = np.array(opts.offset[:] + [0])
        xlen, ylen = glroutines.adjust(b.xlen, b.ylen, w, h)
        offset[0]  = xlen * offset[0] / 2
        offset[1]  = ylen * offset[1] / 2
        offset     = affine.scaleOffsetXform(1, offset)

        # And finally the camera. Typically
        # the Z axis is inferior-superior,
        # so we want that to be pointing up.
        eye     = list(centre)
        eye[1] += 1
        up      = [0, 0, 1]
        camera  = glroutines.lookAt(eye, centre, up)

        # Order is very important!
        xform = affine.concat(offset, scale, camera, rotate)
        np.array(xform, dtype=np.float32)

        self.__viewOffset = offset
        self.__viewScale  = scale
        self.__viewRotate = rotate
        self.__viewCamera = camera
        self.__viewMat    = xform


    def __setViewport(self):
        """Called by :meth:`_draw`. Configures the viewport and calculates
        the model-view trasformation matrix.

        :returns: ``True`` if the viewport was successfully configured,
                  ``False`` otherwise.
        """

        width, height = self.GetScaledSize()
        b             = self.__displayCtx.bounds
        blo           = [b.xlo, b.ylo, b.zlo]
        bhi           = [b.xhi, b.yhi, b.zhi]
        zoom          = self.opts.zoom / 100.0

        if width == 0 or height == 0:
            return False

        # We allow one dimension to be
        # flat, so we can display 2D
        # meshes (e.g. flattened surfaces)
        if np.sum(np.isclose(blo, bhi)) > 1:
            return False

        # Generate the view and projection matrices
        self.__genViewMatrix(width, height)
        projmat, viewport     = glroutines.ortho3D(blo, bhi,
                                                   width, height, zoom)
        self.__projMat        = projmat
        self.__viewport       = viewport
        self.__invViewProjMat = affine.concat(self.__projMat, self.__viewMat)
        self.__invViewProjMat = affine.invert(self.__invViewProjMat)

        return True


    def _draw(self):
        """Draws the scene to the canvas. """

        if self.destroyed:
            return

        if not self._setGLContext():
            return

        opts          = self.opts
        width, height = self.GetScaledSize()

        if not self.__setViewport():
            return

        overlays, globjs = self.getGLObjects()
        rtexs            = []

        if len(overlays) == 0:
            return

        # Draw each overlay to an
        # off-screen render texture.
        for ovl, globj in zip(overlays, globjs):

            rtex    = self.__globjTextures.get(ovl, None)
            display = self.__displayCtx.getDisplay(ovl)

            if not globj.ready():   continue
            if not display.enabled: continue
            if rtex is None:        continue

            log.debug('Drawing %s [%s]', ovl, globj)

            rtexs.append(rtex)
            with rtex.target():
                rtex.shape = width, height
                gl.glViewport(0, 0, width, height)

                # We draw each overlay to a separate offscreen
                # texture, then blend them together (along
                # with the canvas background colour) below.
                # We need to adjust the blend equation so
                # that the clear colour here does not affect
                # the colours from the overlay render, and that
                # alpha blending within the overlay render
                # (specifically w.r.t. tractograms, i.e.
                # transparent streamlines drawn on top of each
                # other) works correctly.
                #
                # I don't understand alpha blending well enough,
                # but these are some good resources:
                #   - http://www.realtimerendering.com/blog/gpus-prefer-premultiplication/
                #   - http://www.adriancourreges.com/blog/2017/05/09/beware-of-transparent-pixels/
                #   - https://developer.nvidia.com/content/alpha-blending-pre-or-not-pre
                #   - https://limnu.com/webgl-blending-youre-probably-wrong/
                glroutines.clear((0, 0, 0, 0))
                gl.glBlendFuncSeparate(gl.GL_SRC_ALPHA,
                                       gl.GL_ONE_MINUS_SRC_ALPHA,
                                       gl.GL_ONE,
                                       gl.GL_ONE_MINUS_SRC_ALPHA)

                globj.preDraw()
                globj.draw3D(self)
                globj.postDraw()

        # draw those off-screen
        # textures to screen
        gl.glViewport(0, 0, width, height)
        glroutines.clear(opts.bgColour)

        texCoords = textures.Texture2D.generateTextureCoords()
        vertices  = np.array([[-1, -1, 0],
                              [-1,  1, 0],
                              [ 1, -1, 0],
                              [ 1, -1, 0],
                              [-1,  1, 0],
                              [ 1,  1, 0]], dtype=np.float32)

        # Use a shader program to sort overlays
        # by depth at every pixel, so they are
        # drawn from farthest to nearest.
        shader = self.__shaders.get(len(rtexs), None)
        if shader is not None:

            for i, rtex in enumerate(rtexs):
                rtex             .bindTexture(int(gl.GL_TEXTURE0) + i * 2)
                rtex.depthTexture.bindTexture(int(gl.GL_TEXTURE0) + i * 2 + 1)

            with shader.loaded(), glroutines.enabled(gl.GL_DEPTH_TEST):
                shader.set(   'bgColour', opts.bgColour)
                shader.setAtt('vertex',   vertices)
                shader.setAtt('texCoord', texCoords)
                shader.draw(gl.GL_TRIANGLES, 0, 6)

            for i, rtex in enumerate(rtexs):
                rtex             .unbindTexture()
                rtex.depthTexture.unbindTexture()

        # GL14 fallback - just draw the textures
        # directly to screen according to overlay
        # order (see getGLObjects - volumes are
        # always drawn last).
        else:
            with glroutines.enabled(gl.GL_DEPTH_TEST):
                for rtex in rtexs:
                    rtex.draw(vertices, useDepth=True)

        if opts.showCursor:
            self.__drawCursor()

        if opts.showLegend:
            self.__drawLegend()

        if opts.showLight:
            self.__drawLight()

        self.getAnnotations().draw3D()


    def __drawCursor(self):
        """Draws three lines at the current :attr:`.DisplayContext.location`.
        """

        opts  = self.opts
        b     = self.__displayCtx.bounds
        pos   = opts.pos
        annot = self.getAnnotations()
        args  = {'colour'    : opts.cursorColour[:3],
                 'lineWidth' : 1,
                 'occlusion' : True}

        annot.line(pos.x, pos.y, pos.x, pos.y, b.zlo, b.zhi, **args)
        annot.line(pos.x, b.ylo, pos.x, b.yhi, pos.z, pos.z, **args)
        annot.line(b.xlo, pos.y, b.xhi, pos.y, pos.z, pos.z, **args)


    def __drawLegend(self):
        """Draws a legend in the bottom left corner of the screen, showing
        anatomical orientation.
        """

        copts      = self.opts
        annot      = self.getAnnotations()
        b          = self.__displayCtx.bounds
        w, h       = self.GetSize()
        xlen, ylen = glroutines.adjust(b.xlen, b.ylen, w, h)

        # A line for each axis
        vertices       = np.zeros((6, 3), dtype=np.float32)
        vertices[0, :] = [-1,  0,  0]
        vertices[1, :] = [ 1,  0,  0]
        vertices[2, :] = [ 0, -1,  0]
        vertices[3, :] = [ 0,  1,  0]
        vertices[4, :] = [ 0,  0, -1]
        vertices[5, :] = [ 0,  0,  1]

        # Each axis line is scaled to
        # 60 pixels, and the legend is
        # offset from the bottom-left
        # corner by twice this amount.
        scale      = [xlen * 30.0 / w] * 3
        offset     = [-0.5 * xlen + 2.0 * scale[0],
                      -0.5 * ylen + 2.0 * scale[1],
                       0]

        # Apply the current camera
        # angle and rotation settings
        # to the legend vertices. Offset
        # anatomical labels off each
        # axis line by a small amount.
        rotation   = affine.decompose(self.viewMatrix)[2]
        labelxform = affine.compose(scale, offset, rotation)
        linexform  = affine.concat(self.projectionMatrix, labelxform)
        labelverts = affine.transform(vertices * 1.2, labelxform)
        lineverts  = affine.transform(vertices,       linexform)
        kwargs     = {
            'colour'    : copts.cursorColour[:3],
            'lineWidth' : 2,
            'occlusion' : False,
            'applyMvp'  : False
        }

        for i in [0, 2, 4]:
            x1, y1, z1 = lineverts[i,     :]
            x2, y2, z2 = lineverts[i + 1, :]
            annot.line(x1, y1, x2, y2, z1, z2, **kwargs)

        canvas = np.array([w, h])
        view   = np.array([xlen, ylen])

        # Draw each label
        for i, label in enumerate(self.__legendLabels):

            # Calculate pixel x/y
            # location for this label
            xx, xy    = canvas * (labelverts[i, :2] + 0.5 * view) / view
            label.pos = (xx, xy)
            label.draw(w, h)


    def __drawLight(self):
        """Draws a representation of the light source. """

        lightPos = self.lightPos
        annot    = self.getAnnotations()
        bounds   = self.__displayCtx.bounds
        centre   = np.array([bounds.xlo + 0.5 * (bounds.xhi - bounds.xlo),
                             bounds.ylo + 0.5 * (bounds.yhi - bounds.ylo),
                             bounds.zlo + 0.5 * (bounds.zhi - bounds.zlo)])

        lx, ly, lz = lightPos
        cx, cy, cz = centre

        annot.point(lx, ly, lz,             colour=(1, 1, 0), lineWidth=3)
        annot.line( lx, ly, cx, cy, lz, cz, colour=(1, 1, 0))



#!/usr/bin/env python
#
# textures.py - Management of OpenGL image textures.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This package is a container for a collection of classes which use OpenGL
textures for various purposes.


.. todo:: There is a lot of duplicate code in the various texture sub-classes.
          This will hopefully be rectified at some stage in the future -
          shared code will be moved into the :class:`.Texture` class.
"""


# All *Texture classes are made available at the
# textures package level due to these imports
from .texture            import Texture                 # noqa
from .texture2d          import Texture2D               # noqa
from .texture2d          import DepthTexture            # noqa
from .texture3d          import Texture3D               # noqa
from .imagetexture       import ImageTexture            # noqa
from .imagetexture       import ImageTexture2D          # noqa
from .imagetexture       import createImageTexture      # noqa
from .imagetexture       import AuxImageTextureManager  # noqa
from .colourmaptexture   import ColourMapTexture        # noqa
from .lookuptabletexture import LookupTableTexture      # noqa
from .selectiontexture   import SelectionTexture2D      # noqa
from .selectiontexture   import SelectionTexture3D      # noqa
from .rendertexture      import RenderTexture           # noqa
from .rendertexture      import GLObjectRenderTexture   # noqa



#!/usr/bin/env python
#
# showcommandline.py - The ShowCommandLineAction class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`ShowCommandLineAction` class, an action
used by the :class:`.CanvasPanel` class. A couple of stand-alone functions
are also defined here:

.. autosummary::
   :nosignatures:

   showCommandLineArgs
   genCommandLineArgs
"""


import wx

import fsl.data.image         as fslimage
import fsleyes_widgets.dialog as fsldlg
import fsleyes.strings        as strings
import fsleyes.parseargs      as parseargs
from . import                    base


class ShowCommandLineAction(base.Action):
    """The :class:`ShowCommandLineAction` class is an :mod:`.action` which is
    used by :class:`.CanvasPanel` instances to generate a FSLeyes command line
    string which can be used to re-create the scene shown in the panel.
    """


    def __init__(self, overlayList, displayCtx, panel):
        """Create a ``ShowCommandLineAction``.

        :arg overlayList: The :class:`.OverlayList`.
        :arg displayCtx:  The :class:`.DisplayContext`.
        :arg panel:       The :class:`.CanvasPanel`.
        """
        base.Action.__init__(
            self, overlayList, displayCtx, self.__showCommandLineArgs)
        self.__panel = panel


    def __showCommandLineArgs(self):
        """Called when this action is executed. Calls the
        :func:`showCommandLineArgs` function.
        """
        showCommandLineArgs(self.overlayList, self.displayCtx, self.__panel)


def showCommandLineArgs(overlayList, displayCtx, canvas):
    """Generates command line arguments which can be used to re-create the
    scene shown on the given :class:`CanvasPanel`, and displays them
    to the user with a :class:`.TextEditDialog`.

    :arg overlayList: A :class:`.OverlayList` .
    :arg displayCtx:  A :class:`.DisplayContext` instance.
    :arg canvas:      A :class:`CanvasPanel` instance.
    """

    unsaved = False

    for o in overlayList:
        if o.dataSource is None:
            unsaved = True
        elif isinstance(o, fslimage.Image) and not o.saveState:
            unsaved = True

    if unsaved:
        msg   = strings.messages[canvas, 'showCommandLineArgs', 'unsaved']
        title = strings.titles[  canvas, 'showCommandLineArgs', 'unsaved']
        wx.MessageBox(msg, title, wx.ICON_EXCLAMATION | wx.OK)
        return

    args = genCommandLineArgs(overlayList, displayCtx, canvas)
    dlg  = fsldlg.TextEditDialog(
        canvas,
        title=strings.messages[  canvas, 'showCommandLineArgs', 'title'],
        message=strings.messages[canvas, 'showCommandLineArgs', 'message'],
        text=' '.join(args),
        icon=wx.ICON_INFORMATION,
        style=(fsldlg.TED_OK        |
               fsldlg.TED_READONLY  |
               fsldlg.TED_MULTILINE |
               fsldlg.TED_COPY      |
               fsldlg.TED_COPY_MESSAGE))

    dlg.CentreOnParent()
    dlg.ShowModal()


def genCommandLineArgs(overlayList, displayCtx, canvas=None):
    """Called by the :func:`showCommandLineArgs` function. Generates
    command line arguments which can be used to re-create the scene
    currently shown on the given :class:`CanvasPanel`.

    :arg overlayList: A :class:`.OverlayList` .
    :arg displayCtx:  A :class:`.DisplayContext` instance.
    :arg canvas:      A :class:`CanvasPanel` instance. If ``None``,
                      scene arguments are not generated.

    :returns:         A list of command line arguments.
    """

    argv = ['fsleyes']

    # Add scene options
    if canvas is not None:
        sceneOpts = canvas.sceneOpts
        argv += parseargs.generateSceneArgs(
            overlayList,
            displayCtx,
            sceneOpts)

    # Add display options for each overlay
    for overlay in displayCtx.getOrderedOverlays():

        fname   = overlay.dataSource
        ovlArgv = parseargs.generateOverlayArgs(overlay,
                                                overlayList,
                                                displayCtx)
        argv   += [fname] + ovlArgv

    return argv



#!/usr/bin/env python
#
# __init__.py - Functions for OpenGL 3.3 rendering.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This package contains modules for rendering various :class:`.GLObject`
types in an OpenGL 3.3 compatible manner.
"""

from .      import gltractogram_funcs  # noqa
from ..gl21 import (gllabel_funcs,
                    gllinevector_funcs,
                    glmask_funcs,
                    glmesh_funcs,
                    glmip_funcs,
                    glrgbvector_funcs,
                    glrgbvolume_funcs,
                    glsh_funcs,
                    gltensor_funcs,
                    glvector_funcs,
                    glvolume_funcs)  # noqa



#!/usr/bin/env python
#
# copyoverlay.py - Action which copies the currently selected overlay.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`CopyOverlayAction`, a global action
which creates a copy of the currently selected overlay.
"""


import numpy as np

import fsl.data.image         as fslimage
import fsl.utils.settings     as fslsettings
import fsl.utils.image.roi    as imgroi
import fsleyes_widgets.dialog as fsldlg
import fsleyes.strings        as strings
from . import                    base


class CopyOverlayAction(base.NeedOverlayAction):
    """The ``CopyOverlayAction`` does as its name suggests - it creates a
    copy of the currently selected overlay.


    .. note:: Currently this action is only capable of copying ``.Image``
              overlays.


    The user is asked to choose between the following options:

     - If the overlay is a 4D ``Image``, should we copy the entire 4D image,
       or extract the current 3D volume?

     - Should we copy all of the ``Image`` data, or create a blank
       (i.e. filled with zeros) ``Image`` of the same dimensions?

     - Should we copy the ``Image`` display properties
       (e.g. :attr:`.Display.overlayType`), or set the display properties
       of the copy to defaults?
    """


    def __init__(self, overlayList, displayCtx, frame):
        """Create a ``CopyOverlayAction``.

        :arg overlayList: The :class:`.OverlayList`.
        :arg displayCtx:  The :class:`.DisplayContext`.
        :arg frame:       The :class:`.FSLeyesFrame`.
        """
        base.NeedOverlayAction.__init__(
            self, overlayList, displayCtx, func=self.__copyOverlay)
        self.__frame = frame


    def __copyOverlay(self):
        """Creates a copy of the currently selected overlay, and inserts it
        into the :class:`.OverlayList`.
        """

        import wx

        overlay = self.displayCtx.getSelectedOverlay()

        if overlay is None:
            return

        # TODO support for other overlay types
        if type(overlay) != fslimage.Image:
            raise RuntimeError('Currently, only {} instances can be '
                               'copied'.format(fslimage.Image.__name__))

        display = self.displayCtx.getDisplay(overlay)

        # We ask the user questions four:
        #  - Copy data, or create an empty (a.k.a. mask) image?
        #  - Copy display settings?
        #  - For 4D, copy 4D, or just the current 3D volume?
        #  - For complex/RGB(A), copy as single channel, or multi-channel?
        #
        # Here we build a list of
        # questions and initial states.
        options = []
        states  = []

        createMaskSetting  = 'fsleyes.actions.copyoverlay.createMask'
        copyDisplaySetting = 'fsleyes.actions.copyoverlay.copyDisplay'
        copy4DSetting      = 'fsleyes.actions.copyoverlay.copy4D'
        copyMultiSetting   = 'fsleyes.actions.copyoverlay.copyMulti'

        createMask  = fslsettings.read(createMaskSetting,  False)
        copyDisplay = fslsettings.read(copyDisplaySetting, False)
        copy4D      = fslsettings.read(copy4DSetting,      False)
        copyMulti   = fslsettings.read(copy4DSetting,      True)
        is4D        = len(overlay.shape) > 3 and overlay.shape[3] > 1
        isMulti     = overlay.iscomplex or overlay.nvals > 1

        options.append(strings.messages['actions.copyoverlay.createMask'])
        states .append(createMask)

        options.append(strings.messages['actions.copyoverlay.copyDisplay'])
        states .append(copyDisplay)

        if is4D:
            options.append(strings.messages['actions.copyoverlay.copy4D'])
            states .append(copy4D)

        if isMulti:
            options.append(strings.messages['actions.copyoverlay.copyMulti'])
            states .append(copyMulti)

        # Ask the user what they want to do
        dlg = fsldlg.CheckBoxMessageDialog(
            self.__frame,
            title=strings.actions[self],
            message='Copy {}'.format(display.name),
            cbMessages=options,
            cbStates=states,
            yesText='OK',
            cancelText='Cancel',
            focus='yes')

        if dlg.ShowModal() != wx.ID_YES:
            return

        createMask            = dlg.CheckBoxState(0)
        copyDisplay           = dlg.CheckBoxState(1)
        if is4D:    copy4D    = dlg.CheckBoxState(2)
        if isMulti: copyMulti = dlg.CheckBoxState(3 if is4D else 2)

        fslsettings            .write(createMaskSetting,  createMask)
        fslsettings            .write(copyDisplaySetting, copyDisplay)
        if is4D:    fslsettings.write(copy4DSetting,      copy4D)
        if isMulti: fslsettings.write(copyMultiSetting,   copyMulti)

        # If the user de-selected copy all channels,
        # ask them which channel they want to copy
        channel = None
        if isMulti and (not copyMulti):
            if overlay.iscomplex:
                choices = ['real', 'imag']
            else:
                choices = ['R' ,'G', 'B', 'A'][:overlay.nvals]

            labels = [strings.choices[self, 'component'][c] for c in choices]
            title  = strings.titles[  'actions.copyoverlay.component']
            msg    = strings.messages['actions.copyoverlay.component']
            dlg    = wx.SingleChoiceDialog(self.__frame,
                                           msg,
                                           title,
                                           choices=labels)

            if dlg.ShowModal() != wx.ID_OK:
                return

            channel = choices[dlg.GetSelection()]

        copyImage(self.overlayList,
                  self.displayCtx,
                  overlay,
                  createMask=createMask,
                  copy4D=copy4D,
                  channel=channel,
                  copyDisplay=copyDisplay)


def copyImage(overlayList,
              displayCtx,
              overlay,
              createMask=False,
              copy4D=True,
              copyDisplay=True,
              name=None,
              roi=None,
              channel=None,
              data=None):
    """Creates a copy of the given :class:`.Image` overlay, and inserts it
    into the :class:`.OverlayList`.

    :arg overlayList: The :class:`.OverlayList`.

    :arg displayCtx:  The :class:`.DisplayContext`.

    :arg overlay:     The :class:`.Image` to be copied.

    :arg createMask:  If ``True``, the copy will be an empty ``Image`` the
                      same shape as the ``overlay``.

    :arg copy4D:      If ``True``, and the ``overlay`` is 4D, the copy will
                      also be 4D. Otherwise, the current 3D voluem is copied.

    :arg copyDisplay: If ``True``, the copy will inherit the display settings
                      of the ``overlay``. Otherwise, the copy will be
                      initialised  with default display settings.

    :arg name:        If provided, will be used as the :attr:`.Display.name`
                      of the copy. Otherwise the copy will be given a name.

    :arg roi:         If provided, the copy will be cropped to the low/high
                      voxel bounds specified in the image. Must be a sequence
                      of tuples, containing the low/high bounds for each voxel
                      dimension. For 4D images, the bounds for the fourth
                      dimension are optional. If ``roi`` specifies more than
                      three dimensions, but ``copy4D is False``, the additional
                      dimensions are ignored.

    :arg channel:     If provided, and if the image is complex or multi-valued
                      (RGB(A)), only this channel is copied. Otherwise the
                      image and data type are copied as-is. For complex images,
                      valid values are ``'real'`` or ``'imag'``; for multi-
                      valued images, valid values are ``'R'``, ``'G'``, ``'B'``
                      or ``'A'``.

    :arg data:        If provided, is used as the image data for the new copy.
                      Must match the shape dictated by the other arguments
                      (i.e. ``copy4D`` and ``roi``). If ``data`` is provided,
                      the ``createMask`` argument is ignored.

    :returns:         The newly created :class:`.Image` object.


    Note that the ``roi`` and ``copy4D`` options do not support images with
    more than four dimensions.
    """

    ovlIdx     = overlayList.index(overlay)
    opts       = displayCtx.getOpts(overlay)
    is4D       = len(overlay.shape) > 3
    isROI      = roi is not None
    copy4D     = copy4D and is4D
    createMask = createMask and (data is None)

    # Initialise the roi indices if one wasn't
    # provided - we will use the indices
    # regardless of whether an ROI was passed
    # in or not
    if roi is None:
        roi = [(0, s) for s in overlay.shape]

    # Adjust the roi to index a
    # specific volume if requested
    if not copy4D:
        roi = list(roi[:3]) + [(i, i + 1) for i in opts.index()[3:]]

    if name is None:
        name = '{}_copy'.format(overlay.name)

    # If an ROI is not specified, we slice
    # the image data, either including all
    # volumes, or the currently selected volume
    if not isROI:
        slc     = tuple(slice(lo, hi) for lo, hi in roi)
        imgdata = overlay[slc]
        xform   = None

    # if an ROI is specified, we use the
    # fsl.utils.image.roi module to generate
    # an ROI and the adjusted voxel-to-world
    # affine
    else:
        roi     = imgroi.roi(overlay, roi)
        imgdata = roi.data
        xform   = roi.voxToWorldMat

    if channel is not None:
        if overlay.iscomplex:
            if   channel == 'real': imgdata = imgdata.real
            elif channel == 'imag': imgdata = imgdata.imag
            else: raise ValueError('Invalid value for channel: '
                                   '{}'.format(channel))
        elif overlay.nvals > 1:
            if channel not in 'RGBA':
                raise ValueError('Invalid value for channel: '
                                   '{}'.format(channel))
            imgdata = imgdata[channel]

    if createMask:
        data = np.zeros(imgdata.shape, dtype=imgdata.dtype)
    elif data is None:
        data = np.copy(imgdata)

    copy = fslimage.Image(data,
                          xform=xform,
                          header=overlay.header,
                          name=name)

    overlayList.insert(ovlIdx + 1, copy)

    # Copy the Display/DisplayOpts settings
    if copyDisplay:

        # Don't override the name
        # that we set above
        dispexcl = ('name',)

        # And don't clobber the transform, and related,
        # properties, as it is (typically) automatically
        # controlled via the DisplayContext.displaySpace
        optexcl = ('transform', 'bounds')

        copyDisplayProperties(displayCtx,
                              overlay,
                              copy,
                              displayExclude=dispexcl,
                              optExclude=optexcl)

    return copy


def copyDisplayProperties(displayCtx,
                          src,
                          dest,
                          displayExclude=None,
                          optExclude=None,
                          displayArgs=None,
                          optArgs=None):
    """Copies all properties from the ``src`` :class:`.Display` and
    :class:`.DisplayOpts` instances to the ``dest`` instances.

    :arg displayCtx:     The :class:`.DisplayContext`

    :arg displayExclude: Collection of :class:`.Display` properties which
                         should not be copied

    :arg optExclude:     Collection of :class:`.DisplayOpts` properties which
                         should not be copied

    :arg displayArgs:    Values to be used instead of the ``src`` ``Display``
                         values.

    :arg optArgs:        Values to be used instead of the ``src``
                         ``DisplayOpts`` values.
    """

    if displayExclude is None: displayExclude = []
    if optExclude     is None: optExclude     = []
    if displayArgs    is None: displayArgs    = {}
    if optArgs        is None: optArgs        = {}

    # copy the Display properties first, as
    # they may affect the DisplayOpts type
    srcDisplay  = displayCtx.getDisplay(src)
    destDisplay = displayCtx.getDisplay(dest)

    for prop in srcDisplay.getAllProperties()[0]:

        if prop in displayExclude:
            continue

        if (not srcDisplay .propertyIsEnabled(prop)) or \
           (not destDisplay.propertyIsEnabled(prop)):
            continue

        val = displayArgs.get(prop, getattr(srcDisplay, prop))

        # Check that the overlay type of the old
        # overlay is valid for the new overlay
        if prop == 'overlayType':
            choices = destDisplay.getProp('overlayType').getChoices(destDisplay)
            if val not in choices:
                continue

        setattr(destDisplay, prop, val)

    # And after the Display has been configured
    # copy the DisplayOpts settings.
    srcOpts  = displayCtx.getOpts(src)
    destOpts = displayCtx.getOpts(dest)

    for prop in srcOpts.getAllProperties()[0]:

        if prop in optExclude:
            continue

        # The source and destination opts
        # instances may be different types
        if not hasattr(destOpts, prop):
            continue

        if (not srcOpts .propertyIsEnabled(prop)) or \
           (not destOpts.propertyIsEnabled(prop)):
            continue

        val = optArgs.get(prop, getattr(srcOpts, prop))
        setattr(destOpts, prop, val)



#!/usr/bin/env python
#
# loadcolourmap.py - Load a colour map file.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`LoadColourMapAction`, which allows the
user to load a new colour map.
"""

import            logging
import            os
import os.path as op

import fsl.utils.settings           as fslsettings
import fsleyes_widgets.utils.status as status
import fsleyes.strings              as strings
import fsleyes.colourmaps           as fslcmap
from . import                          base


log = logging.getLogger(__name__)


class LoadColourMapAction(base.Action):
    """The ``LoadColourMapAction`` allows the user to select a colour
    map file and give it a name.

    The loaded colour map is then registered with the :mod:`.colourmaps`
    module. The user is also given the option to permanently save the
    loaded colour map in *FSLeyes*.
    """


    def __init__(self, overlayList, displayCtx):
        """Create a ``LoadColourMapAction``.

        :arg overlayList: The :class:`.OverlayList`.
        :arg displayCtx:  The :class:`.DisplayContext`.
        """
        base.Action.__init__(
            self, overlayList, displayCtx, self.__loadColourMap)


    def __loadColourMap(self):
        """This method does the following:

        1. Prompts the user to select a colour map file

        2. Prompts the user to name the new colour map.

        3. Registers the colour map with the :mod:`.colourmaps` module.

        4. Asks the user if they want the colour map installed, and installs
           it if they do.
        """

        import wx

        app = wx.GetApp()

        # Get the most recent colour map
        # directory if there is one
        loadDir = fslsettings.read('fsleyes.loadcolourmapdir', os.getcwd())

        # prompt the user to choose a colour map file
        dlg = wx.FileDialog(
            app.GetTopWindow(),
            defaultDir=loadDir,
            message=strings.messages[self, 'loadcmap'],
            style=wx.FD_OPEN)

        if dlg.ShowModal() != wx.ID_OK:
            return

        # prompt the user to choose  a name for the colour
        # map (using the filename prefix as the default)
        cmapFile = dlg.GetPath()
        cmapDir  = op.dirname(cmapFile)
        cmapName = op.splitext(op.basename(cmapFile))[0]

        cmapNameMsg   = strings.messages[self, 'namecmap']
        cmapNameTitle = strings.titles[  self, 'namecmap']

        while True:

            dlg = wx.TextEntryDialog(
                app.GetTopWindow(),
                cmapNameMsg,
                cmapNameTitle,
                cmapName)

            if dlg.ShowModal() != wx.ID_OK:
                return

            cmapName = dlg.GetValue()
            cmapKey  = fslcmap.makeValidMapKey(cmapName)

            # a colour map with the specified name already exists
            if fslcmap.isColourMapRegistered(cmapKey):
                cmapNameMsg = strings.messages[self, 'alreadyinstalled']
                continue

            break

        # register the selected colour map file
        fslcmap.registerColourMap(cmapFile,
                                  self.overlayList,
                                  self.displayCtx,
                                  key=cmapKey,
                                  name=cmapName)

        # Save the directory for next time
        fslsettings.write('fsleyes.loadcolourmapdir', cmapDir)

        # ask the user if they want to install
        # the colour map for future use
        dlg = wx.MessageDialog(
            app.GetTopWindow(),
            caption=strings.titles[  self, 'installcmap'],
            message=strings.messages[self, 'installcmap'],
            style=wx.YES_NO)

        if dlg.ShowModal() != wx.ID_YES:
            return

        # install the colour map
        etitle = strings.titles[  self, 'installerror']
        emsg   = strings.messages[self, 'installerror']

        with status.reportIfError(title=etitle, msg=emsg, raiseError=False):
            fslcmap.installColourMap(cmapKey)



#!/usr/bin/env python
#
# glmesh_funcs.py - OpenGL 2.1 functions used by the GLMesh class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides functions which are used by the :class:`.GLMesh`
class to render :class:`.Mesh` overlays in an OpenGL 2.1 compatible
manner.
"""


import fsleyes.gl.shaders as shaders


def compileShaders(self):
    """Loads the ``glmesh`` vertex/fragment shader source and creates
    :class:`.GLSLShader` instance(s).
    """

    if self.threedee:

        flatVertSrc = shaders.getVertexShader(  'glmesh_3d_flat')
        flatFragSrc = shaders.getFragmentShader('glmesh_3d_flat')
        dataVertSrc = shaders.getVertexShader(  'glmesh_3d_data')
        dataFragSrc = shaders.getFragmentShader('glmesh_3d_data')

        self.flatShader = shaders.GLSLShader(flatVertSrc, flatFragSrc)
        self.dataShader = shaders.GLSLShader(dataVertSrc, dataFragSrc)

    else:

        flatVertSrc    = shaders.getVertexShader(  'glmesh_2d_flat')
        flatFragSrc    = shaders.getFragmentShader('glmesh_2d_flat')
        dataVertSrc    = shaders.getVertexShader(  'glmesh_2d_data')
        dataFragSrc    = shaders.getFragmentShader('glmesh_2d_data')
        xsectcpVertSrc = shaders.getVertexShader(  'glmesh_2d_crosssection_clipplane')
        xsectcpFragSrc = shaders.getFragmentShader('glmesh_2d_crosssection_clipplane')
        xsectblVertSrc = shaders.getVertexShader(  'glmesh_2d_crosssection_blit')
        xsectblFragSrc = shaders.getFragmentShader('glmesh_2d_crosssection_blit')

        self.dataShader    = shaders.GLSLShader(dataVertSrc,    dataFragSrc)
        self.flatShader    = shaders.GLSLShader(flatVertSrc,    flatFragSrc)
        self.xsectcpShader = shaders.GLSLShader(xsectcpVertSrc, xsectcpFragSrc)
        self.xsectblShader = shaders.GLSLShader(xsectblVertSrc, xsectblFragSrc)


def updateShaderState(self, **kwargs):
    """Updates the shader program according to the current :class:`.MeshOpts``
    configuration.
    """

    dopts      = self.opts
    dshader    = self.dataShader
    fshader    = self.flatShader
    xscpshader = self.xsectcpShader
    xsblshader = self.xsectblShader

    with dshader.loaded():
        dshader.set('cmap',           0)
        dshader.set('negCmap',        1)
        dshader.set('useNegCmap',     kwargs['useNegCmap'])
        dshader.set('cmapXform',      kwargs['cmapXform'])
        dshader.set('flatColour',     kwargs['flatColour'])
        dshader.set('invertClip',     dopts.invertClipping)
        dshader.set('discardClipped', dopts.discardClipped)
        dshader.set('modulateAlpha',  dopts.modulateAlpha)
        dshader.set('modScale',       kwargs['modScale'])
        dshader.set('modOffset',      kwargs['modOffset'])
        dshader.set('clipLow',        dopts.clippingRange.xlo)
        dshader.set('clipHigh',       dopts.clippingRange.xhi)

        if self.threedee:
            dshader.setIndices(self.indices)
            dshader.setAtt('vertex', self.vertices)
            dshader.setAtt('normal', self.normals)

            vdata = self.getVertexData('vertex')
            mdata = self.getVertexData('modulate')

            # if modulate data is not set,
            # we use the vertex data
            if mdata is None:
                mdata = vdata

            if vdata is not None: dshader.setAtt('vertexData',   vdata)
            if mdata is not None: dshader.setAtt('modulateData', mdata)

    with fshader.loaded():
        fshader.set('colour', kwargs['flatColour'])
        if self.threedee:
            fshader.setAtt('vertex', self.vertices)
            fshader.setAtt('normal', self.normals)
            fshader.setIndices(self.indices)

    if not self.threedee:
        with xscpshader.loaded():
            xscpshader.setAtt('vertex', self.vertices)
            xscpshader.setIndices(      self.indices)
        with xsblshader.loaded():
            xsblshader.set('colour', kwargs['flatColour'])



#!/usr/bin/env python
#
# newimage.py - The NewImageAction class
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`NewImageAction` class, an :class:`.Action`
which allows the user to create a new empty :class:`.Image` overlay.
"""


import            wx
import wx.grid as wxgrid

import numpy   as np
import nibabel as nib

import fsl.transform.affine      as fslaffine
import fsl.data.constants        as constants
import fsl.data.image            as fslimage
import fsleyes_widgets.floatspin as floatspin
import fsleyes.strings           as strings
from . import                       base


class NewImageAction(base.Action):
    """The ``NewImageAction`` class allows the user to create a new
    :class:`.Image`.  When invoked, it displays a :class:`NewImageDialog`
    prompting the user to select the properties of the new image, and then
    creates a new image accordingly.
    """

    def __init__(self, overlayList, displayCtx, frame):
        """Create a ``CopyOverlayAction``.

        :arg overlayList: The :class:`.OverlayList`.
        :arg displayCtx:  The :class:`.DisplayContext`.
        :arg frame:       The :class:`.FSLeyesFrame`.
        """
        base.Action.__init__(self, overlayList, displayCtx, self.__newImage)
        self.__frame = frame


    def __newImage(self):
        """Displays a :class:`NewImageDialog`, then creates a new
        :class:`.Image`, and adds it to the :class:`.OverlayList`.

        If the currently selected overlay is a :class:`.Nifti`, the
        :class:`NewImageDialog` is initialised to the properties of
        the selected overlay.
        """

        ovl = self.displayCtx.getSelectedOverlay()

        if ovl is not None and isinstance(ovl, fslimage.Nifti):
            shape     = ovl.shape[:3]
            pixdim    = ovl.pixdim[:3]
            dtype     = ovl.dtype
            affine    = ovl.voxToWorldMat
            xyzUnits  = ovl.xyzUnits
            timeUnits = ovl.timeUnits

            # adjust pixdims in case there
            # are inversions (e.g. l/r flip)
            pixdim = pixdim * np.sign(fslaffine.decompose(affine)[0])
        else:
            shape     = None
            pixdim    = None
            dtype     = None
            affine    = np.eye(4)
            xyzUnits  = None
            timeUnits = None

        dlg = NewImageDialog(self.__frame,
                             shape=shape,
                             pixdim=pixdim,
                             affine=affine,
                             dtype=dtype)

        if dlg.ShowModal() != wx.ID_OK:
            return

        img = newImage(dlg.shape,
                       dlg.pixdim,
                       dlg.dtype,
                       dlg.affine,
                       xyzUnits=xyzUnits,
                       timeUnits=timeUnits,
                       name='new')
        self.overlayList.append(img)
        self.displayCtx.selectOverlay(img)


def newImage(shape,
             pixdim,
             dtype,
             affine,
             xyzUnits=constants.NIFTI_UNITS_MM,
             timeUnits=constants.NIFTI_UNITS_SEC,
             name='new'):
    """Create a new :class:`.Image` with the specified properties.

    :arg shape:      Tuple containing the image shape
    :arg pixdim:     Tuple containing the image pixdims
    :arg dtype:      ``numpy`` ``dtype``
    :arg affine:     ``(4, 4)`` ``numpy`` array specifying the voxel-to-world
                     affine
    :arg xyzUnits:   Spatial units
    :arg timeUnits:  Temporal units
    """
    data = np.zeros(shape, dtype=dtype)
    img  = nib.Nifti2Image(data, affine)
    img.header.set_zooms(np.abs(pixdim))
    img.header.set_xyzt_units(xyzUnits, timeUnits)
    return fslimage.Image(img, name=name)


class NewImageDialog(wx.Dialog):
    """The ``NewImageDialog`` displays a collection of widgets allowing the
    user to select the data type, shape, dimensions (pixdims), and voxel-to-
    world affine.
    """

    def __init__(self, parent, shape, pixdim, affine, dtype):
        """Create a ``NewImageDialog``.

        :arg parent: ``wx`` parent object
        :arg shape:  Tuple of three initial shape values
        :arg pixdim: Tuple of three initial pixdim values
        :arg affine: Initial affine, assumed to be a ``(4, 4)`` ``numpy``
                     array
        :arg dtype:  Initial ``numpy`` dtype. Must be one of ``float32``,
                     ``float64``, ``uint8``, ``int16``, or ``int32``.
        """

        if shape  is None: shape  = (100, 100, 100)
        if pixdim is None: pixdim = (1, 1, 1)
        if affine is None: affine = np.eye(4)
        if dtype  is None: dtype  = np.float32

        dtypeLabels = ['float',    'uchar',  'sshort', 'sint',   'double']
        dtypeValues = [np.float32, np.uint8, np.int16, np.int32, np.float64]
        dtypeLabels = [strings.labels[self, l] for l in dtypeLabels]

        if dtype in dtypeValues: dtype = dtypeValues.index(dtype)
        else:                    dtype = 0

        self.__dtypeValues = dtypeValues

        wx.Dialog.__init__(self,
                           parent,
                           title=strings.titles[self],
                           style=wx.DEFAULT_DIALOG_STYLE)

        dtypeLabel  = strings.labels[self, 'dtype']
        shapeLabel  = strings.labels[self, 'shape']
        pixdimLabel = strings.labels[self, 'pixdim']
        affineLabel = strings.labels[self, 'affine']
        linkLabel   = strings.labels[self, 'link']
        okLabel     = strings.labels[self, 'ok']
        cancelLabel = strings.labels[self, 'cancel']

        self.__dtypeLabel  = wx.StaticText(self, label=dtypeLabel)
        self.__shapeLabel  = wx.StaticText(self, label=shapeLabel)
        self.__pixdimLabel = wx.StaticText(self, label=pixdimLabel)
        self.__affineLabel = wx.StaticText(self, label=affineLabel)
        self.__xLabel      = wx.StaticText(self, label='X')
        self.__yLabel      = wx.StaticText(self, label='Y')
        self.__zLabel      = wx.StaticText(self, label='Z')
        self.__dtype       = wx.Choice(self, choices=dtypeLabels)
        self.__ok          = wx.Button(self, id=wx.ID_OK,     label=okLabel)
        self.__cancel      = wx.Button(self, id=wx.ID_CANCEL, label=cancelLabel)

        self.__dtype.SetSelection(dtype)

        shapewidgets  = []
        pixdimwidgets = []

        for i in range(3):
            shapew  = floatspin.FloatSpinCtrl(self,
                                              minValue=1,
                                              maxValue=2 ** 64 - 1,
                                              increment=1,
                                              value=shape[i],
                                              style=floatspin.FSC_INTEGER)
            pixdimw = floatspin.FloatSpinCtrl(self,
                                              minValue=-100,
                                              maxValue=100,
                                              increment=0.5,
                                              value=pixdim[i])

            shapewidgets .append(shapew)
            pixdimwidgets.append(pixdimw)

        self.__shapex  = shapewidgets[ 0]
        self.__shapey  = shapewidgets[ 1]
        self.__shapez  = shapewidgets[ 2]
        self.__pixdimx = pixdimwidgets[0]
        self.__pixdimy = pixdimwidgets[1]
        self.__pixdimz = pixdimwidgets[2]
        self.__link    = wx.CheckBox(self, label=linkLabel)
        self.__link.SetValue(True)

        self.__affine = wxgrid.Grid(self)
        self.__affine.SetDefaultEditor(wxgrid.GridCellFloatEditor(-1, 2))
        self.__affine.CreateGrid(4, 4)
        self.__affine.HideRowLabels()
        self.__affine.HideColLabels()
        for i in range(4):
            self.__affine.SetColFormatFloat(i, -1, 2)
        self.affine = affine

        self.__mainSizer      = wx.BoxSizer(wx.VERTICAL)
        self.__dtypeSizer     = wx.BoxSizer(wx.HORIZONTAL)
        self.__affineSizer    = wx.BoxSizer(wx.HORIZONTAL)
        self.__affineLblSizer = wx.BoxSizer(wx.HORIZONTAL)
        self.__shapepixSizer  = wx.BoxSizer(wx.HORIZONTAL)
        self.__shapepixGrid   = wx.FlexGridSizer(4, 3, 0, 0)
        self.__buttonSizer    = wx.BoxSizer(wx.HORIZONTAL)

        self.__dtypeSizer.Add((1, 1), proportion=1)
        self.__dtypeSizer.Add(self.__dtypeLabel)
        self.__dtypeSizer.Add(self.__dtype)
        self.__dtypeSizer.Add((1, 1), proportion=1)

        self.__shapepixGrid .Add((1, 1))
        self.__shapepixGrid .Add(self.__shapeLabel)
        self.__shapepixGrid .Add(self.__pixdimLabel)
        self.__shapepixGrid .Add(self.__xLabel)
        self.__shapepixGrid .Add(self.__shapex)
        self.__shapepixGrid .Add(self.__pixdimx)
        self.__shapepixGrid .Add(self.__yLabel)
        self.__shapepixGrid .Add(self.__shapey)
        self.__shapepixGrid .Add(self.__pixdimy)
        self.__shapepixGrid .Add(self.__zLabel)
        self.__shapepixGrid .Add(self.__shapez)
        self.__shapepixGrid .Add(self.__pixdimz)
        self.__shapepixSizer.Add((1, 1), proportion=1)
        self.__shapepixSizer.Add(self.__shapepixGrid)
        self.__shapepixSizer.Add((1, 1), proportion=1)

        self.__affineSizer.Add((1, 1), proportion=1)
        self.__affineSizer.Add(self.__affine)
        self.__affineSizer.Add((1, 1), proportion=1)

        self.__affineLblSizer.Add(self.__affineLabel)
        self.__affineLblSizer.Add((10, 1))
        self.__affineLblSizer.Add(self.__link)

        self.__buttonSizer.Add((1, 1), proportion=1)
        self.__buttonSizer.Add(self.__ok)
        self.__buttonSizer.Add(self.__cancel)
        self.__buttonSizer.Add((1, 1), proportion=1)

        szargs = {'flag'   : wx.EXPAND | wx.LEFT | wx.RIGHT,
                  'border' : 5}

        self.__mainSizer.Add((1, 10))
        self.__mainSizer.Add(self.__dtypeSizer, **szargs)
        self.__mainSizer.Add((1, 10))
        self.__mainSizer.Add(self.__shapepixSizer, **szargs)
        self.__mainSizer.Add((1, 10))
        self.__mainSizer.Add(self.__affineLblSizer, **szargs)
        self.__mainSizer.Add((1, 10))
        self.__mainSizer.Add(self.__affineSizer, **szargs)
        self.__mainSizer.Add((1, 10))
        self.__mainSizer.Add(self.__buttonSizer, **szargs)
        self.__mainSizer.Add((1, 10))

        self.SetSizer(self.__mainSizer)
        self.Layout()
        self.Fit()
        self.__ok.SetFocus()

        self.__pixdimx.Bind(floatspin.EVT_FLOATSPIN,      self.__onPixdim)
        self.__pixdimy.Bind(floatspin.EVT_FLOATSPIN,      self.__onPixdim)
        self.__pixdimz.Bind(floatspin.EVT_FLOATSPIN,      self.__onPixdim)
        self.__affine .Bind(wxgrid.EVT_GRID_CELL_CHANGED, self.__onAffine)
        self.__link   .Bind(wx.EVT_CHECKBOX,              self.__onLink)


    def __onLink(self, ev):
        """Called when the link checkbox changes. If linking is enabled,
        calls :meth:`__onAffine`.
        """
        if not self.link:
            return

        # affine takes precedence
        self.__onAffine(ev)



    def __onPixdim(self, ev):
        """Called when any pixdim widget changes. If linking is enabled,
        reconstructs the affine with the new pixdim values.
        """
        if not self.link:
            return

        # If we construct an affine with
        # scales == 0, we get explosions
        if any([p == 0 for p in self.pixdim]):
            return

        offsets, rotations = fslaffine.decompose(self.affine)[1:]
        self.affine = fslaffine.compose(self.pixdim, offsets, rotations)


    def __onAffine(self, ev):
        """Called when the affine changes. If linking is enabled, updates
        the pixdim values from the new affine.
        """
        if not self.link:
            return

        scales = fslaffine.decompose(self.affine)[0]
        self.__pixdimx.SetValue(scales[0])
        self.__pixdimy.SetValue(scales[1])
        self.__pixdimz.SetValue(scales[2])


    @property
    def linkWidget(self):
        """Return a reference to the link widget. """
        return self.__link


    @property
    def shapeWidgets(self):
        """Return a reference to the three shape widgets. """
        return (self.__shapex, self.__shapey, self.__shapez)


    @property
    def pixdimWidgets(self):
        """Return a reference to the three pixdim widgets. """
        return (self.__pixdimx, self.__pixdimy, self.__pixdimz)


    @property
    def dtypeWidget(self):
        """Return a reference to the dtype widget. """
        return self.__dtype


    @property
    def affineWidget(self):
        """Return a reference to the affine widget. """
        return self.__affine


    @property
    def ok(self):
        """Return a reference to the ok button. """
        return self.__ok


    @property
    def cancel(self):
        """Return a reference to the cancel button. """
        return self.__cancel


    @property
    def link(self):
        """Return a tuple containing the current affine-dimension link checkbox
        value.
        """
        return self.__link.GetValue()


    @property
    def shape(self):
        """Return a tuple containing the current shape values. """
        return (self.__shapex.GetValue(),
                self.__shapey.GetValue(),
                self.__shapez.GetValue())


    @property
    def pixdim(self):
        """Return a tuple containing the current pixdim values. """
        return (self.__pixdimx.GetValue(),
                self.__pixdimy.GetValue(),
                self.__pixdimz.GetValue())


    @property
    def dtype(self):
        """Return the currently selected data type, as a ``numpy.dtype``. """
        return self.__dtypeValues[self.__dtype.GetSelection()]


    @property
    def affine(self):
        """Return the current content of the affine grid, as a ``numpy``
        array of shape ``(4, 4)``.
        """
        aff = np.zeros((4, 4), dtype=np.float64)
        for i in range(4):
            for j in range(4):
                aff[i, j] = float(self.__affine.GetCellValue(i, j))
        return aff


    @affine.setter
    def affine(self, aff):
        """Set the current contents of the affine grid to ``aff``, assumed
        to be a ``(4, 4)`` ``numpy`` array.
        """
        for i in range(4):
            for j in range(4):
                val = '{:0.2f}'.format(aff[i, j])
                self.__affine.SetCellValue((i, j), val)



#!/usr/bin/env python
#
# resources.py - Simple manager for shared OpenGL resources.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module implements a simple API for managing shared OpenGL resources.
Some OpenGL resources (e.g. textures) take up a lot of memory so it makes
sense to share these resources where possible, instead of creating and
maintaining multiple copies. The API defined in this module consists of the
following functions:


.. autosummary::
   :nosignatures:

   exists
   get
   set
   delete


On creation, resources must be given a unique name, referred to as a
``key``. Subsequent accesses to the resource are performed by specifying this
key. As an example, let's say that we have a :class:`.Image` called
``myImage``::


    import fsleyes.gl.resources as glresources
    import fsleyes.gl.textures  as gltextures
    import fsl.data.image       as fslimage

    image = fslimage.Image('data.nii.gz')

We wish to create an :class:`.ImageTexture` which can be shared by multiple
users. All users of this texture can use the :func:`get` function to access
the texture. The first call to :func:`get` will result in the texture being
created, whereas subsequent calls will return a reference to the existing
texture, and will increase its reference count::

    texture = glresources.get(
        'myTexture',
         gltextures.ImageTexture,
         'myTexture',
         image,
         interp=gl.GL_LINEAR)


.. note:: Here, we have used ``'myTexture'`` as the resource key. In practice,
          you will need to use something that is guaranteed to be unique
          throughout your application.


When a user of the texture no longer needs the texture, it must call the
:func:`delete` method. Calls to :func:`delete` will decrement the reference
count; when this count reaches zero, the texture will be destroyed::


    glresources.delete('myTexture')


.. note:: This module was written for managing OpenGL :class:`.Texture`
          objects, but can actually be used with any type - the only
          requirement is that the type defines a method called ``destroy``,
          which performs any required clean-up operations.
"""

import logging


log = logging.getLogger(__name__)


def exists(key):
    """Returns ``True`` if a resource with the specified key exists, ``False``
    otherwise.
    """
    return key in _resources


def get(key, createFunc=None, *args, **kwargs):
    """Return a reference to the resource wiuh the specified key.

    If no resource with the given key exists, and ``createFunc`` is not
    ``None``, the resource is created, registered, and returned. If
    the resource does not exist, and ``createFunc`` is ``None``, a
    :exc:`KeyError` is raised.

    :arg key:        Unique resource identifier.
    :arg createFunc: If the resource does not exist, and this argument
                     is provided, it will be called to create the resource.

    All other positional and keyword arguments will be passed through to
    the ``createFunc``.
    """

    r = _resources.get(key, None)

    if r is None and createFunc is None:
        raise KeyError(f'Resource {key} does not exist')

    if r is not None:
        r.refcount += 1

        log.debug('Resource %s reference  count '
                  'increased to %s', str(key), r.refcount)

        return r.resource

    if createFunc is not None:
        return set(key, createFunc(*args, **kwargs))


def set(key, resource, overwrite=False):
    """Create a new resource, or update an existing one.

    :arg key:       Unique resource identifier.

    :arg resource:  The resource itself.

    :arg overwrite: If ``False`` (the default), and a resource with
                    the specified ``key`` already exists, a :exc:`KeyError`
                    is raised. Otherwise, it is assumed that a resource with
                    the specified ``key`` exists - the existing resource is
                    replaced with the specified ``resource``.
    """

    if (not overwrite) and (key in _resources):
        raise KeyError(f'Resource {key} already exists')

    if not overwrite:
        log.debug('Adding resource {}'.format(str(key)))

        r               = _Resource(key, resource)
        r.refcount     += 1
        _resources[key] = r

        log.debug('Resource %s reference count '
                  'increased to %s', str(key), r.refcount)

    else:
        log.debug('Updating resource %s', str(key))

        _resources[key].resource = resource

    return resource


def delete(key):
    """Decrements the reference count of the resource with the specified key.
    When the resource reference count reaches ``0``, the ``destroy`` method
    is called on the resource.

    :arg key: Unique resource identifier.
    """

    r           = _resources[key]
    r.refcount -= 1

    log.debug('Resource %s reference count '
              'decreased to %s', str(key), r.refcount)

    if r.refcount <= 0:

        log.debug('Destroying resource %s', str(key))

        _resources.pop(key)
        r.resource.destroy()


class _Resource:
    """Internal type which is used to encapsulate a resource, and the
    number of active references to that resources. The following attributes
    are available on a ``_Resource``:

    ============ ============================================================
    ``key``      The unique resource key.
    ``resource`` The resource itself.
    ``refcount`` Number of references to the resource (initialised to ``0``).
    ============ ============================================================
    """

    def __init__(self, key, resource):
        """Create a ``_Resource``.

        :arg key:      The unique resource key.
        :arg resource: The resource itself.
        """
        self.key      = key
        self.resource = resource
        self.refcount = 0


_resources = {}
"""A dictionary containing ``{key : _Resource}`` mappings for all resources
that exist.
"""



#!/usr/bin/env python
#
# selectiontexture.py - The SelectionTexture class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`SelectionTexture2D` and
:class:`SelectionTexture3D` classes, :class:`.Texture` types which can be used
to store :class:`.Selection` instances.

The ``SelectionTexture2D/3D`` classes are used by the :class:`.VoxelSelection`
annotation to display the contents of a :class:`.Selection` instance.
"""


import logging

import numpy     as np
import OpenGL.GL as gl

import fsl.transform.affine as affine
from . import                  texture2d
from . import                  texture3d


log = logging.getLogger(__name__)


class SelectionTextureBase:
    """Base class shared by the :class:`SelectionTexture2D` and
    :class:`SelectionTexture3D`. Manages updates from the :class:`.Selection`
    object.
    """

    def __init__(self, selection):
        """
        This method must be called *after* :meth:`.Texture.__init__`.
        """
        self.__selection = selection
        selection.register(self.name, self.__selectionChanged)
        self.__selectionChanged(init=True)


    @property
    def selection(self):
        """Returns a reference to the :class:`.Selection` object. """
        return self.__selection


    def destroy(self):
        """Must be called when this ``SelectionTextureBase`` is no longer
        needed. Removes the listener on the :attr:`.Selection.selection`
        property.
        """
        self.__selection.deregister(self.name)
        self.__selection = None


    def __selectionChanged(self, *a, **kwa):
        """Called when the :attr:`.Selection.selection` changes. Updates
        the texture data via the :meth:`.Texture.doPatch` method.
        """

        init             = kwa.pop('init', False)
        old, new, offset = self.__selection.getLastChange()
        shape            = self.__selection.shape

        def prepare(data, oldShape=None):
            data = self.shapeData(data, oldShape=oldShape)
            return (data * 255).astype(np.uint8)

        if init or (new is None):
            data = prepare(self.__selection.getSelection(), shape)
            self.set(data=data)
        else:
            data   = prepare(new)
            offset = affine.transform(offset, self.texCoordXform(shape))
            self.doPatch(data, offset)


class SelectionTexture3D(texture3d.Texture3D, SelectionTextureBase):
    """The ``SelectionTexture3D`` class is a :class:`.Texture3D` which can be
    used to store a :class:`.Selection` instance.  The ``Selection`` image
    array is stored as a single channel 3D texture, which is updated whenever
    the :attr:`.Selection.selection` property changes - updates are managed by
    the :class:`SelectionTextureBase` class.
    """


    def __init__(self, name, selection):
        """Create a ``SelectionTexture3D``.

        :arg name:      A unique name for this ``SelectionTexture3D``.
        :arg selection: The :class:`.Selection` instance.
        """

        texture3d.Texture3D .__init__(self, name, nvals=1)
        SelectionTextureBase.__init__(self, selection)


    def destroy(self):
        """Must be called when this ``SelectionTexture3D`` is no longer needed.
        Calls the :meth:`.Texture.destroy` method, and removes the listener
        on the :attr:`.Selection.selection` property.
        """
        texture3d.Texture3D .destroy(self)
        SelectionTextureBase.destroy(self)


class SelectionTexture2D(texture2d.Texture2D, SelectionTextureBase):
    """The ``SelectionTexture2D`` class is a :class:`.Texture2D` which can be
    used to store a :class:`.Selection` instance.  The ``Selection`` image
    array is stored as a single channel 2D texture, which is updated whenever
    the :attr:`.Selection.selection` property changes - updates are managed by
    the :class:`SelectionTextureBase` class..
    """


    def __init__(self, name, selection):
        """Create a ``SelectionTexture2D``.

        :arg name:      A unique name for this ``SelectionTexture2D``.
        :arg selection: The :class:`.Selection` instance.
        """

        texture2d.Texture2D .__init__(self, name, nvals=1)
        SelectionTextureBase.__init__(self, selection)


    def destroy(self):
        """Must be called when this ``SelectionTexture2D`` is no longer needed.
        Calls the :meth:`.Texture.destroy` method, and removes the listener
        on the :attr:`.Selection.selection` property.
        """
        texture2d.Texture2D .destroy(self)
        SelectionTextureBase.destroy(self)



#!/usr/bin/env python
#
# glvector_funcs.py - Functions used by glrgbvector_funcs and
#                     gllinevector_funcs.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module contains logic for managing vertex and fragment shader programs
used for rendering :class:`.GLRGBVector` and :class:`.GLLineVector` instances.
These functions are used by the :mod:`.gl21.glrgbvector_funcs` and
:mod:`.gl21.gllinevector_funcs` modules.
"""


import numpy as np

import fsl.data.constants   as constants
import fsl.transform.affine as affine
import fsleyes.gl.shaders   as shaders


def compileShaders(self, vertShader):
    """Compiles the vertex/fragment shader programs (by creating a
    :class:`.GLSLShader` instance).

    If the :attr:`.VectorOpts.colourImage` property is set, the ``glvolume``
    fragment shader is used. Otherwise, the ``glvector`` fragment shader
    is used.
    """

    if self.shader is not None:
        self.shader.destroy()

    opts                = self.opts
    useVolumeFragShader = opts.colourImage is not None

    self.useVolumeFragShader = useVolumeFragShader

    if useVolumeFragShader: fragShader = 'glvolume'
    else:                   fragShader = 'glvector'

    vertSrc = shaders.getVertexShader(  vertShader)
    fragSrc = shaders.getFragmentShader(fragShader)

    return shaders.GLSLShader(vertSrc, fragSrc)


def updateShaderState(self, useSpline=False):
    """Updates the state of the vector vertex and fragment shader. The fragment
    shader may be either the ``glvolume`` or the ``glvector`` shader.
    The shader is assumed to be loaded.
    """

    changed           = False
    opts              = self.opts
    shader            = self.shader
    imageShape        = self.vectorImage.shape[:3]
    modLow,  modHigh  = self.getModulateRange()
    clipLow, clipHigh = self.getClippingRange()

    if opts.modulateImage is None: modShape  = [1, 1, 1]
    else:                          modShape  = opts.modulateImage.shape[:3]
    if opts.clipImage     is None: clipShape = [1, 1, 1]
    else:                          clipShape = opts.clipImage.shape[:3]

    modMode     = {'brightness' : 0,
                   'alpha'      : 1}[opts.modulateMode]

    clipXform   = self.getAuxTextureXform('clip')
    colourXform = self.getAuxTextureXform('colour')
    modXform    = self.getAuxTextureXform('modulate')

    changed |= self.shader.set('clipCoordXform',   clipXform)
    changed |= self.shader.set('colourCoordXform', colourXform)
    changed |= self.shader.set('modCoordXform',    modXform)

    if self.useVolumeFragShader:

        voxValXform   = self.colourTexture.voxValXform
        img2CmapXform = affine.concat(
            self.cmapTexture.getCoordinateTransform(),
            voxValXform)

        changed |= shader.set('clipTexture',      1)
        changed |= shader.set('imageTexture',     2)
        changed |= shader.set('colourTexture',    3)
        changed |= shader.set('negColourTexture', 3)
        changed |= shader.set('img2CmapXform',    img2CmapXform)
        changed |= shader.set('imageShape',       imageShape)
        changed |= shader.set('imageIsClip',      False)
        changed |= shader.set('useNegCmap',       False)
        changed |= shader.set('useSpline',        useSpline)
        changed |= shader.set('clipLow',          clipLow)
        changed |= shader.set('clipHigh',         clipHigh)
        changed |= shader.set('invertClip',       False)

    else:
        # If we are displaying an RGB24 image
        # as a vector, we map uint8 [0, 255] to
        # [-1, 1]. The integer values will be
        # automatically normalised to [0, 1],
        # so we transform from that range.
        if self.vectorImage.niftiDataType == constants.NIFTI_DT_RGB24:
            voxValXform = affine.scaleOffsetXform(2, -1)

        # Otherwise, if it's floating point,
        # it will not be normalised.
        else:
            voxValXform = self.imageTexture.voxValXform

        colours, colourXform = opts.getVectorColours()

        changed |= shader.set('modulateTexture', 0)
        changed |= shader.set('clipTexture',     1)
        changed |= shader.set('vectorTexture',   4)
        changed |= shader.set('xColour',         colours[0])
        changed |= shader.set('yColour',         colours[1])
        changed |= shader.set('zColour',         colours[2])
        changed |= shader.set('colourXform',     colourXform)
        changed |= shader.set('voxValXform',     voxValXform)
        changed |= shader.set('imageShape',      imageShape)
        changed |= shader.set('modImageShape',   modShape)
        changed |= shader.set('clipImageShape',  clipShape)
        changed |= shader.set('clipLow',         clipLow)
        changed |= shader.set('clipHigh',        clipHigh)
        changed |= shader.set('modLow',          modLow)
        changed |= shader.set('modHigh',         modHigh)
        changed |= shader.set('useSpline',       useSpline)
        changed |= shader.set('modulateMode',    modMode)

    return changed



#!/usr/bin/env python
#
# annotationpanel.py - The AnnotationPanel
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module contains the :class:`AnnotationPanel` class, a FSLeyes control
panel which allows the user to annotate the canvases of an
:class:`.OrthoPanel`.
"""


import copy

import wx

import fsleyes_widgets.elistbox                      as elb
import fsleyes_props                                 as props
import fsleyes_widgets.widgetlist                    as widgetlist
import fsleyes_widgets.bitmapradio                   as bmpradio
import fsleyes.actions                               as actions
import fsleyes.strings                               as strings
import fsleyes.icons                                 as fslicons
import fsleyes.tooltips                              as tooltips
import fsleyes.views.orthopanel                      as orthopanel
import fsleyes.controls.controlpanel                 as ctrlpanel
import fsleyes.plugins.profiles.orthoannotateprofile as annotprofile
import fsleyes.plugins.tools.saveannotations         as saveannotations
import fsleyes.gl.annotations                        as annotations


class AnnotationPanel(ctrlpanel.ControlPanel):
    """The ``AnnotationPanel`` contains controls allowing the user to add
    annotations to the canvases in an :class:`.OrthoPanel`. The user is able
    to add points, lines, rectangles, ellipses, and text to any canvas, and
    can adjust the properties (e.g. colour, thickness) of each annotation.

    When a user selects an annotation to add, the
    :class:`.OrthoAnnotateProfile` is enabled on the ``OrthoPanel``, which
    provides user interaction.

    The :class:`.SaveAnnotationsAction` and :class:`.LoadAnnotationsAction`
    actions, for saving/loading annotations to/from files, are bound to
    buttons in the ``AnnotationPanel``.
    """


    # These properties are used to synchronise property
    # values between the ortho edit profile, and newly
    # added or selected annotation objects
    colour        = copy.copy(annotprofile.OrthoAnnotateProfile.colour)
    fontSize      = copy.copy(annotprofile.OrthoAnnotateProfile.fontSize)
    lineWidth     = copy.copy(annotprofile.OrthoAnnotateProfile.lineWidth)
    filled        = copy.copy(annotprofile.OrthoAnnotateProfile.filled)
    border        = copy.copy(annotprofile.OrthoAnnotateProfile.border)
    honourZLimits = copy.copy(annotprofile.OrthoAnnotateProfile.honourZLimits)
    alpha         = copy.copy(annotprofile.OrthoAnnotateProfile.alpha)


    # Just used as a convenience
    DISPLAY_PROPERTIES = ['colour', 'fontSize', 'lineWidth', 'filled',
                          'border', 'honourZLimits', 'alpha']


    @staticmethod
    def supportedViews():
        """Overrides :meth:`.ControlMixin.supportedViews`. The
        ``CropImagePanel`` is only intended to be added to
        :class:`.OrthoPanel` views.
        """
        return [orthopanel.OrthoPanel]


    @staticmethod
    def profileCls():
        """Returns the :class:`.OrthoAnnotateProfile` class, which needs to be
        activated in conjunction with the ``AnnotationPanel``.
        """
        return annotprofile.OrthoAnnotateProfile


    @staticmethod
    def defaultLayout():
        """Returns a dictionary containing layout settings to be passed to
        :class:`.ViewPanel.togglePanel`.
        """
        return {'location' : wx.LEFT}


    def __init__(self, parent, overlayList, displayCtx, ortho):
        """Create an ``AnnotationPanel``.

        :arg parent:      The :mod:`wx` parent object.
        :arg overlayList: The :class:`.OverlayList` instance.
        :arg displayCtx:  The :class:`.DisplayContext` instance.
        :arg ortho:       The :class:`.OrthoPanel` instance.
        """
        ctrlpanel.ControlPanel.__init__(
            self, parent, overlayList, displayCtx, ortho)
        self.__ortho = ortho

        # load/save actions are bound to buttons below
        self.loadAnnotations = saveannotations.LoadAnnotationsAction(
            overlayList, displayCtx, ortho)
        self.saveAnnotations = saveannotations.SaveAnnotationsAction(
            overlayList, displayCtx, ortho)

        # The interface comprises a list of annotations,
        # a set of controls below the list, a column of
        # buttons down the side, and a row of buttons
        # along the bottom
        self.__mainSizer       = wx.BoxSizer(wx.VERTICAL)
        self.__listSizer       = wx.BoxSizer(wx.HORIZONTAL)
        self.__sideButtonSizer = wx.BoxSizer(wx.VERTICAL)

        # Listbox containing all annotations that
        # have been added to all canvases, sorted
        # by creation time. We add our own buttons
        # to remove/reorder anontations below
        self.__annotList = elb.EditableListBox(
            self, style=(elb.ELB_NO_ADD    |
                         elb.ELB_NO_REMOVE |
                         elb.ELB_NO_MOVE   |
                         elb.ELB_REVERSE))

        # Buttons running down the side of the list,
        # for loading/saving annotations, and for
        # removing/reordering annotations from/within
        # the list
        loadSpec = actions.ActionButton(
            'loadAnnotations',
            icon=fslicons.findImageFile('folder16'),
            tooltip=tooltips.actions[self, 'saveAnnotations'])
        saveSpec = actions.ActionButton(
            'saveAnnotations',
            icon=fslicons.findImageFile('floppydisk16'),
            tooltip=tooltips.actions[self, 'loadAnnotations'])

        self.__loadButton   = props.buildGUI(self, self, loadSpec)
        self.__saveButton   = props.buildGUI(self, self, saveSpec)
        self.__upButton     = wx.Button(self, style=wx.BU_EXACTFIT)
        self.__downButton   = wx.Button(self, style=wx.BU_EXACTFIT)
        self.__removeButton = wx.Button(self, style=wx.BU_EXACTFIT)

        upicon     = wx.Bitmap(fslicons.findImageFile('up16'))
        downicon   = wx.Bitmap(fslicons.findImageFile('down16'))
        removeicon = wx.Bitmap(fslicons.findImageFile('remove16'))

        self.__upButton    .SetBitmapLabel(upicon)
        self.__downButton  .SetBitmapLabel(downicon)
        self.__removeButton.SetBitmapLabel(removeicon)

        self.__sideButtonSizer.Add(self.__loadButton)
        self.__sideButtonSizer.Add(self.__saveButton)
        self.__sideButtonSizer.Add(self.__upButton)
        self.__sideButtonSizer.Add(self.__downButton)
        self.__sideButtonSizer.Add(self.__removeButton)
        self.__sideButtonSizer.Add((1, 1), proportion=1)

        # Stores a reference to the most recently
        # selected annotation object, as we need
        # to bind/unbind property listeners on it
        self.__selected = None

        # Widgets that can be used to set display
        # properties for newly added annotations,
        # and for the existing currently selected
        # annotation in the listbox
        self.__widgets       = widgetlist.WidgetList(self, minHeight=24)
        self.__colour        = props.makeWidget(self.__widgets, self, 'colour')
        self.__filled        = props.makeWidget(self.__widgets, self, 'filled')
        self.__border        = props.makeWidget(self.__widgets, self, 'border')
        self.__honourZLimits = props.makeWidget(self.__widgets, self,
                                                'honourZLimits')
        self.__fontSize      = props.makeWidget(self.__widgets, self,
                                                'fontSize', slider=False)
        self.__lineWidth     = props.makeWidget(self.__widgets, self,
                                                'lineWidth', slider=False)
        self.__alpha         = props.makeWidget(self.__widgets, self,
                                                'alpha', spin=False)

        self.__widgets.AddWidget(self.__colour,
                                 strings.labels[self, 'colour'])
        self.__widgets.AddWidget(self.__alpha,
                                 strings.labels[self, 'alpha'])
        self.__widgets.AddWidget(self.__fontSize,
                                 strings.labels[self, 'fontSize'])
        self.__widgets.AddWidget(self.__lineWidth,
                                 strings.labels[self, 'lineWidth'])
        self.__widgets.AddWidget(self.__filled,
                                 strings.labels[self, 'filled'])
        self.__widgets.AddWidget(self.__border,
                                 strings.labels[self, 'border'])
        self.__widgets.AddWidget(self.__honourZLimits,
                                 strings.labels[self, 'honourZLimits'])

        # Radio box buttons allowing the user to
        # select an annotation type to draw
        self.__annotOptions = bmpradio.BitmapRadioBox(
            self, style=bmpradio.BMPRADIO_ALLOW_DESELECTED)

        icons = {
            'text'    : [fslicons.loadBitmap('textAnnotationHighlight24'),
                         fslicons.loadBitmap('textAnnotation24')],
            'rect'    : [fslicons.loadBitmap('rectAnnotationHighlight24'),
                         fslicons.loadBitmap('rectAnnotation24')],
            'line'    : [fslicons.loadBitmap('lineAnnotationHighlight24'),
                         fslicons.loadBitmap('lineAnnotation24')],
            'arrow'   : [fslicons.loadBitmap('arrowAnnotationHighlight24'),
                         fslicons.loadBitmap('arrowAnnotation24')],
            'point'   : [fslicons.loadBitmap('pointAnnotationHighlight24'),
                         fslicons.loadBitmap('pointAnnotation24')],
            'ellipse' : [fslicons.loadBitmap('ellipseAnnotationHighlight24'),
                         fslicons.loadBitmap('ellipseAnnotation24')],
        }

        for option, icons in icons.items():
            self.__annotOptions.AddChoice(*icons, clientData=option)

        self.__listSizer.Add(self.__sideButtonSizer)
        self.__listSizer.Add(self.__annotList, flag=wx.EXPAND, proportion=1)

        self.__mainSizer.Add(self.__listSizer,    flag=wx.EXPAND, proportion=1)
        self.__mainSizer.Add(self.__widgets,      flag=wx.EXPAND)
        self.__mainSizer.Add(self.__annotOptions, flag=wx.EXPAND)

        self.SetSizer(self.__mainSizer)

        self.__upButton    .Bind(wx.EVT_BUTTON, self.__onMoveUp)
        self.__downButton  .Bind(wx.EVT_BUTTON, self.__onMoveDown)
        self.__removeButton.Bind(wx.EVT_BUTTON, self.__onRemove)

        self.__annotOptions.Bind(bmpradio.EVT_BITMAP_RADIO_EVENT,
                                 self.__onAnnotOption)

        self.__annotList.Bind(elb.EVT_ELB_SELECT_EVENT,
                              self.__annotListItemSelected)

        for propName in self.DISPLAY_PROPERTIES:
            self.addListener(propName, self.name,
                             self.__displayPropertyChanged)

        for canvas in ortho.getGLCanvases():
            canvas.getAnnotations().addListener(
                'annotations', self.name, self.__annotationsChanged)

        # populate annotations listbox
        self.__annotationsChanged()


    def destroy(self):
        """Must be called when this ``AnnoationPanel`` is no longer needed.
        Removes property listeners and clears references.
        """
        super().destroy()
        for canvas in self.__ortho.getGLCanvases():
            canvas.getAnnotations().removeListener('annotations', self.name)
        self.__annotListItemSelected()
        self.__ortho = None


    def __annotationsChanged(self, *a):
        """Called when the :attr:`.Annotations.annotations` list of any
        :class:`.SliceCanvas` changes. Makes sure that the contents of the
        annotations list box is up to date, and selects the most recently
        added annotation.
        """

        # This is a little awkward. Each XYZ canvas has its own
        # separate list of annotation objects, but we display
        # all objects in a single list, and the user is able to
        # arbitrarily re-order the objects, independently of
        # which canvas they are from (even though reordering
        # two objects from different canvases has no effect on
        # how they are drawn).
        #
        # The __onMove method keeps the ordering of the GUI list
        # and canvas lists in sync. But this method (which is
        # typically called when new annotations are added) just
        # clears and re-creates the GUI list from the canvas
        # lists. This keeps the code simple, but does mean that
        # new annotations will appear after the last annotation
        # *from the same canvas*, meaning that they may appear
        # in the middle of the GUI list.
        alist     = self.__annotList
        allAnnots = []
        for canvas in self.__ortho.getGLCanvases():
            allAnnots.extend(canvas.getAnnotations().annotations)

        listAnnots = alist.GetData()
        alist.Clear()

        for obj in allAnnots:
            canvas = 'XYZ'[obj.annot.canvas.opts.zax]
            if isinstance(obj, annotations.TextAnnotation):
                name = obj.text
            else:
                name = strings.labels[self, obj]
            alist.Append('[{}] {}'.format(canvas, name), obj)

        # select the most recently added annotation
        # (we just take the first one which was not
        # originally in the GUI list) and bind it
        # to the display settings widgets
        for i, obj in enumerate(allAnnots):
            if obj not in listAnnots:
                alist.SetSelection(i)
                self.__annotListItemSelected(obj=obj)
                break


    def __displayPropertyChanged(self, *a):
        """Called when any display property changes. Refreshes all canvases.
        """
        if self.__selected is not None:
            self.__ortho.Refresh()


    def __onRemove(self, ev):
        """Called when the "remove item" button is pushed.  Removes the
        item from the list box widget, and from the corresponding
        :attr:`.Annotations.annotations` list.
        """

        idx = self.__annotList.GetSelection()
        if idx == wx.NOT_FOUND:
            return

        obj   = self.__annotList.GetItemData(idx)
        annot = obj.annot

        self.__annotList.Delete(idx)

        with props.suppress(annot, 'annotations'):
            annot.dequeue(obj, hold=True, fixed=False)

        obj.annot.canvas.Refresh()

        if self.__annotList.GetCount() > 0:
            self.__annotList.SetSelection(max((0, idx - 1)))
            idx = self.__annotList.GetSelection()
            obj = self.__annotList.GetItemData(idx)
            self.__annotListItemSelected(obj=obj)


    def __onMoveUp(self, ev):
        """Called when the "move item up" button is pushed. """
        self.__onMove(-1)


    def __onMoveDown(self, ev):
        """Called when the "move item down" button is pushed. """
        self.__onMove(1)


    def __onMove(self, offset):
        """Called when one of the move buttons is pushed. Moves the item in the
        list box, then syncs the new item order on the corresponding
        :attr:`.Annotations.annotations` list.
        """
        alist = self.__annotList
        idx   = alist.GetSelection()

        if idx == wx.NOT_FOUND:
            return

        obj   = alist.GetItemData(idx)
        annot = obj.annot

        alist.MoveItem(offset)

        # Sync the order of the canvas annotations
        # list with the new order in the GUI list
        with props.suppress(annot, 'annotations'):
            allobjs = [o for o in alist.GetData() if o.annot is annot]
            annot.annotations[:] = list(allobjs)
        annot.canvas.Refresh()


    def __annotListItemSelected(self, ev=None, obj=None):
        """Called when an item is selected in the annotations listbox, or
        programmatically at various times. If the properties of an
        :class:`.AnnotationObject` are bound to the display settings widgets,
        they are unbound. If an annotation has been selected in the annotations
        list box, the display settings widgets are bound to its properties.
        """

        if ev is not None:
            obj = ev.data

        prev            = self.__selected
        self.__selected = obj

        for propName in self.DISPLAY_PROPERTIES:
            if prev is not None and hasattr(prev, propName):
                self.unbindProps(propName, prev)
            if obj is not None and hasattr(obj, propName):
                self.bindProps(propName, obj)


    def __onAnnotOption(self, ev):
        """Called when the user selects an annnotation type to draw. Changes
        the :class:`.OrthoAnnotateProfile` mode.
        """
        self.__annotList.ClearSelection()
        self.__annotListItemSelected()

        profile = self.__ortho.currentProfile

        if ev.value:
            profile.mode = ev.clientData
            for propName in self.DISPLAY_PROPERTIES:
                profile.bindProps(propName, self)
        else:
            profile.mode = 'nav'
